"""
Code to support recording to a readable and executable Python script.

TODO:
    - Support for dictionaries?

"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008-2015, Enthought, Inc.
# License: BSD Style.

import warnings
import sys
PY_VER = sys.version_info[0]
import __builtin__

from traits.api import (HasTraits, List, Str, Dict, Bool,
        Unicode, Property, Int, Instance)
from traits.util.camel_case import camel_case_to_python


################################################################################
# `_RegistryData` class.
################################################################################
class _RegistryData(HasTraits):
    # Object's script ID
    script_id = Property(Str)

    # Path to object in object hierarchy.
    path = Property(Str)

    # Parent data for this object if any.
    parent_data = Instance('_RegistryData', allow_none=True)

    # The name of the trait on the parent which is this object.
    trait_name_on_parent = Str('')

    # List of traits we are listening for on this object.
    names = List(Str)

    # Nested recordable instances on the object.
    sub_recordables = List(Str)

    # List of traits that are lists.
    list_names = List(Str)

    _script_id = Str('')

    ######################################################################
    # Non-public interface.
    ######################################################################
    def _get_path(self):
        pdata = self.parent_data
        path = ''
        if pdata is not None:
            pid = pdata.script_id
            ppath = pdata.path
            tnop = self.trait_name_on_parent
            if '[' in tnop:
                # If the object is a nested object through an iterator,
                # we instantiate it and don't refer to it through the
                # path, this makes scripting convenient.
                if len(ppath) == 0:
                    path = pid + '.' + tnop
                else:
                    path = ppath + '.' + tnop
            else:
                path = ppath + '.' + tnop

        return path

    def _get_script_id(self):
        sid = self._script_id
        if len(sid) == 0:
            pdata = self.parent_data
            sid = pdata.script_id + '.' + self.trait_name_on_parent
        return sid

    def _set_script_id(self, id):
        self._script_id = id



################################################################################
# `RecorderError` class.
################################################################################
class RecorderError(Exception):
    pass


################################################################################
# `Recorder` class.
################################################################################
class Recorder(HasTraits):

    # The lines of code recorded.
    lines = List(Str)

    # Are we recording or not?
    recording = Bool(False, desc='if script recording is enabled or not')

    # The Python script we have recorded so far.  This is just a
    # convenience trait for the `get_code()` method.
    script = Property(Unicode)

    ########################################
    # Private traits.

    # Dict used to store information on objects registered.  It stores a
    # unique name for the object and its path in the object hierarchy
    # traversed.
    _registry = Dict

    # Reverse registry with keys as script_id and object as value.
    _reverse_registry = Dict

    # A mapping to generate unique names for objects.  The key is the
    # name used (which is something derived from the class name of the
    # object) and the value is an integer describing the number of times
    # that variable name has been used earlier.
    _name_map = Dict(Str, Int)

    # A list of special reserved script IDs.  This is handy when you
    # want a particular object to have an easy to read script ID and not
    # the default one based on its class name.  This leads to slightly
    # easier to read scripts.
    _special_ids = List

    # What are the known names in the script?  By known names we mean
    # names which are actually bound to objects.
    _known_ids = List(Str)

    # The known types in the namespace.
    _known_types = List(Str)

    # A guard to check if we are currently in a recorded function call,
    # in which case we don't want to do any recording.
    _in_function = Bool(False)

    ######################################################################
    # `Recorder` interface.
    ######################################################################
    def record(self, code):
        """Record a string to be stored to the output file.

        Parameters:
        -----------

        code - A string of text.
        """
        if self.recording and not self._in_function:
            lines = self.lines
            # Analyze the code and add extra code if needed.
            self._analyze_code(code)
            # Add the code.
            lines.append(code)

    def register(self, object, parent=None, trait_name_on_parent='',
                 ignore=None, known=False, script_id=None):
        """Register an object with the recorder.  This sets up the
        object for recording.

        By default all traits (except those starting and ending with
        '_') are recorded.  For attributes that are themselves
        recordable, one may mark traits with a 'record' metadata as
        follows:

        - If metadata `record=False` is set, the nested object will not be
          recorded.

        - If `record=True`, then that object is also recorded if it is
          not `None`.

        If the object is a list or dict that is marked with
        `record=True`, the list is itself not listened to for changes
        but all its contents are registered.

        If the `object` has a trait named `recorder` then this recorder
        instance will be set to it if possible.

        Parameters:
        -----------

        object : Instance(HasTraits)
            The object to register in the registry.

        parent : Instance(HasTraits)
            An optional parent object in which `object` is contained

        trait_name_on_parent : str
            An optional trait name of the `object` in the `parent`.

        ignore : list(str)
            An optional list of trait names on the `object` to be
            ignored.

        known : bool
            Optional specification if the `object` id is known on the
            interpreter.  This is needed if you are manually injecting
            code to define/create an object.

        script_id : str
            Optionally specify a script_id to use for this object.  It
            is not guaranteed that this ID will be used since it may
            already be in use.
        """
        registry = self._registry

        # Do nothing if the object is already registered.
        if object in registry:
            return

        # When parent is specified the trait_name_on_parent must also be.
        if parent is not None:
            assert len(trait_name_on_parent) > 0

        if ignore is None:
            ignore = []

        if isinstance(object, HasTraits):
            # Always ignore these.
            ignore.extend(['trait_added', 'trait_modified'])

            sub_recordables = object.traits(record=True).keys()
            # Find all the trait names we must ignore.
            ignore.extend(object.traits(record=False).keys())
            # The traits to listen for.
            tnames = [t for t in object.trait_names()
                      if not t.startswith('_') and not t.endswith('_') \
                         and t not in ignore]
            # Find all list traits.
            trts = object.traits()
            list_names = []
            for t in tnames:
                tt = trts[t].trait_type
                if hasattr(tt, 'default_value_type') and \
                        tt.default_value_type == 5:
                    list_names.append(t)
        else:
            # No traits, so we can't do much.
            sub_recordables = []
            tnames = []
            list_names = []

        # Setup the registry data.

        # If a script id is supplied try and use it.
        sid = ''
        if script_id is not None:
            r_registry = self._reverse_registry
            while script_id in r_registry:
                script_id = '%s1'%script_id
            sid = script_id
            # Add the chosen id to special_id list.
            self._special_ids.append(sid)

        if parent is None:
            pdata = None
            if len(sid) == 0:
                sid = self._get_unique_name(object)
        else:
            pdata = self._get_registry_data(parent)
            tnop = trait_name_on_parent
            if '[' in tnop:
                # If the object is a nested object through an iterator,
                # we instantiate it and don't refer to it through the
                # path, this makes scripting convenient.
                sid = self._get_unique_name(object)

        # Register the object with the data.
        data = _RegistryData(script_id=sid,
                             parent_data=pdata,
                             trait_name_on_parent=trait_name_on_parent,
                             names=tnames,
                             sub_recordables=sub_recordables,
                             list_names=list_names)
        registry[object] = data

        # Now get the script id of the object -- note that if sid is ''
        # above then the script_id is computed from that of the parent.
        sid = data.script_id
        # Setup reverse registry so we can get the object from the
        # script_id.
        self._reverse_registry[sid] = object

        # Record the script_id if the known argument is explicitly set to
        # True.
        if known:
            self._known_ids.append(sid)

        # Try and set the recorder attribute if necessary.
        if hasattr(object, 'recorder'):
            try:
                object.recorder = self
            except Exception as e:
                msg = "Cannot set 'recorder' trait of object %r: "\
                      "%s"%(object, e)
                warnings.warn(msg, warnings.RuntimeWarning)

        if isinstance(object, HasTraits):
            # Add handler for lists.
            for name in list_names:
                object.on_trait_change(self._list_items_listner,
                                       '%s_items'%name)

            # Register all sub-recordables.
            for name in sub_recordables:
                obj = getattr(object, name)
                if isinstance(obj, list):
                    # Don't register the object itself but register its
                    # children.
                    for i, child in enumerate(obj):
                        attr = '%s[%d]'%(name, i)
                        self.register(child, parent=object,
                                      trait_name_on_parent=attr)
                elif obj is not None:
                    self.register(obj, parent=object,
                                  trait_name_on_parent=name)
                # Listen for changes to the trait itself so the newly
                # assigned object can also be listened to.
                object.on_trait_change(self._object_changed_handler, name)
            # Now add listner for the object itself.
            object.on_trait_change(self._listner, tnames)

    def unregister(self, object):
        """Unregister the given object from the recorder.  This inverts
        the logic of the `register(...)` method.
        """
        registry = self._registry
        # Do nothing if the object isn't registered.
        if object not in registry:
            return

        data = registry[object]

        # Try and unset the recorder attribute if necessary.
        if hasattr(object, 'recorder'):
            try:
                object.recorder = None
            except Exception as e:
                msg = "Cannot unset 'recorder' trait of object %r:"\
                      "%s"%(object, e)
                warnings.warn(msg, warnings.RuntimeWarning)

        if isinstance(object, HasTraits):
            # Remove all list_items handlers.
            for name in data.list_names:
                object.on_trait_change(self._list_items_listner,
                                       '%s_items'%name, remove=True)

            # Unregister all sub-recordables.
            for name in data.sub_recordables:
                obj = getattr(object, name)
                if isinstance(obj, list):
                    # Unregister the children.
                    for i, child in enumerate(obj):
                        self.unregister(child)
                elif obj is not None:
                    self.unregister(obj)
                # Remove the trait handler for trait assignments.
                object.on_trait_change(self._object_changed_handler,
                                       name, remove=True)
            # Now remove listner for the object itself.
            object.on_trait_change(self._listner, data.names, remove=True)

        # Remove the object data from the registry etc.
        if data.script_id in self._known_ids:
            self._known_ids.remove(data.script_id)
        del self._reverse_registry[data.script_id]
        del registry[object]

    def save(self, file):
        """Save the recorded lines to the given file.  It does not close
        the file.
        """
        if PY_VER == 3:
            file.write(self.get_code())
        else:
            file.write(unicode(self.get_code(), encoding='utf-8'))
        file.flush()

    def record_function(self, func, args, kw):
        """Record a function call given the function and its
        arguments."""
        if self.recording and not self._in_function:
            # Record the function name and arguments.
            call_str = self._function_as_string(func, args, kw)
            # Call the function.
            try:
                self._in_function = True
                result = func(*args, **kw)
            finally:
                self._in_function = False

            # Register the result if it is not None.
            if func.__name__ == '__init__':
                f_self = args[0]
                code = self._import_class_string(f_self.__class__)
                self.lines.append(code)
                return_str = self._registry.get(f_self).script_id
            else:
                return_str = self._return_as_string(result)
            if len(return_str) > 0:
                self.lines.append('%s = %s'%(return_str, call_str))
            else:
                self.lines.append('%s'%(call_str))
        else:
            result = func(*args, **kw)
        return result

    def ui_save(self):
        """Save recording to file, pop up a UI dialog to find out where
        and close the file when done.
        """
        from pyface.api import FileDialog, OK
        wildcard = 'Python files (*.py)|*.py|' + FileDialog.WILDCARD_ALL
        dialog = FileDialog(title='Save Script',
                            action='save as', wildcard=wildcard
                            )
        if dialog.open() == OK:
            fname = dialog.path
            f = open(fname, 'w')
            self.save(f)
            f.close()

    def clear(self):
        """Clears all previous recorded state and unregisters all
        registered objects."""
        # First unregister any registered objects.
        registry = self._registry
        while len(registry) > 0:
            self.unregister(registry.keys()[0])
        # Clear the various lists.
        self.lines[:] = []
        self._registry.clear()
        self._known_ids[:] = []
        self._name_map.clear()
        self._reverse_registry.clear()
        self._known_types[:] = []
        self._special_ids[:] = []

    def get_code(self):
        """Returns the recorded lines as a string of printable code."""
        return '\n'.join(self.lines) + '\n'

    def is_registered(self, object):
        """Returns True if the given object is registered with the
        recorder."""
        return object in self._registry

    def get_script_id(self, object):
        """Returns the script_id of a registered object.  Useful when
        you want to manually add a record statement."""
        return self._get_registry_data(object).script_id

    def get_object_path(self, object):
        """Returns the path in the object hierarchy of a registered
        object.  Useful for debugging."""
        return self._get_registry_data(object).path

    def write_script_id_in_namespace(self, script_id):
        """If a script_id is not known in the current script's namespace,
        this sets it using the path of the object or actually
        instantiating it.  If this is not possible (since the script_id
        matches no existing object), nothing is recorded but the
        framework is notified that the particular script_id is available
        in the namespace.  This is useful when you want to inject code
        in the namespace to create a particular object.
        """
        if not self.recording:
            return
        known_ids = self._known_ids
        if script_id not in known_ids:
            obj = self._reverse_registry.get(script_id)
            # Add the ID to the known_ids.
            known_ids.append(script_id)
            if obj is not None:
                data = self._registry.get(obj)
                result = ''
                if len(data.path) > 0:
                    # Record code for instantiation of object.
                    result = '%s = %s'%(script_id, data.path)
                else:
                    # This is not the best thing to do but better than
                    # nothing.
                    result = self._import_class_string(obj.__class__)
                    cls = obj.__class__.__name__
                    mod = obj.__module__
                    result += '\n%s = %s()'%(script_id, cls)

                if len(result) > 0:
                    self.lines.extend(result.split('\n'))

    ######################################################################
    # Non-public interface.
    ######################################################################
    def _get_unique_name(self, obj):
        """Return a unique object name (a string).  Note that this does
        not cache the object, so if called with the same object 3 times
        you'll get three different names.
        """
        cname = obj.__class__.__name__
        nm = self._name_map
        result = ''
        builtin = False
        if cname in __builtin__.__dict__:
            builtin = True
            if hasattr(obj, '__name__'):
                cname = obj.__name__
        else:
            cname = camel_case_to_python(cname)

        special_ids = self._special_ids
        while len(result) == 0 or result in special_ids:
            if cname in nm:
                id = nm[cname] + 1
                nm[cname] = id
                result = '%s%d'%(cname, id)
            else:
                nm[cname] = 0
                # The first id doesn't need a number if it isn't builtin.
                if builtin:
                    result = '%s0'%(cname)
                else:
                    result = cname
        return result

    def _get_registry_data(self, object):
        """Get the data for an object from registry."""
        data = self._registry.get(object)
        if data is None:
            msg = "Recorder: Can't get script_id since object %s not registered"
            raise RecorderError(msg%(object))
        return data

    def _listner(self, object, name, old, new):
        """The listner for trait changes on an object.

        This is called by child listners or when any of the recordable
        object's traits change when recording to a script is enabled.

        Parameters:
        -----------

        object : Object which has changed.

        name : extended name of attribute that changed.

        old : Old value.

        new : New value.

        """
        if self.recording and not self._in_function:
            new_repr = repr(new)
            sid = self._get_registry_data(object).script_id
            if len(sid) == 0:
                msg = '%s = %r'%(name, new)
            else:
                msg = '%s.%s = %r'%(sid, name, new)
            if new_repr.startswith('<') and new_repr.endswith('>'):
                self.record('# ' + msg)
            else:
                self.record(msg)

    def _list_items_listner(self, object, name, old, event):
        """The listner for *_items on list traits of the object.
        """
        # Set the path of registered objects in the modified list and
        # all their children.  This is done by unregistering the object
        # and re-registering them.  This is slow but.
        registry = self._registry
        sid = registry.get(object).script_id
        trait_name = name[:-6]
        items = getattr(object, trait_name)
        for (i, item) in enumerate(items):
            if item in registry:
                data = registry.get(item)
                tnop = data.trait_name_on_parent
                if len(tnop) > 0:
                    data.trait_name_on_parent = '%s[%d]'%(trait_name, i)

        # Record the change.
        if self.recording and not self._in_function:
            index = event.index
            removed = event.removed
            added = event.added
            nr = len(removed)
            slice = '[%d:%d]'%(index, index + nr)
            na = len(added)
            rhs = [self._object_as_string(item) for item in added]
            rhs = ', '.join(rhs)
            obj = '%s.%s'%(sid, name[:-6])
            msg = '%s%s = [%s]'%(obj, slice, rhs)
            self.record(msg)

    def _object_changed_handler(self, object, name, old, new):
        """Called when a child recordable object has been reassigned."""
        registry = self._registry
        if old is not None:
            if old in registry:
                self.unregister(old)
        if new is not None:
            if new not in registry:
                self.register(new, parent=object,
                              trait_name_on_parent=name)

    def _get_script(self):
        return self.get_code()

    def _analyze_code(self, code):
        """Analyze the code and return extra code if needed.
        """
        known_ids = self._known_ids
        lhs = ''
        try:
            lhs = code.split()[0]
        except IndexError:
            pass

        if '.' in lhs:
            ob_name = lhs.split('.')[0]
            self.write_script_id_in_namespace(ob_name)

    def _function_as_string(self, func, args, kw):
        """Return a string representing the function call."""
        func_name = func.__name__
        func_code = func.func_code
        # Even if func is really a decorated method it never shows up as
        # a bound or unbound method here, so we have to inspect the
        # argument names to figure out if this is a method or function.
        if func_code.co_argcount > 0 and \
           func_code.co_varnames[0] == 'self':
            # This is a method, the first argument is bound to self.
            f_self = args[0]
            # Convert the remaining arguments to strings.
            argl = [self._object_as_string(arg) for arg in args[1:]]

            # If this is __init__ we special case it.
            if func_name == '__init__':
                # Register the object.
                self.register(f_self, known=True)
                func_name = f_self.__class__.__name__
            else:
                sid = self._object_as_string(f_self)
                func_name = '%s.%s'%(sid, func_name)
        else:
            argl = [self._object_as_string(arg) for arg in args]

        # Convert the keyword args.
        kwl = ['%s=%s'%(key, self._object_as_string(value))
               for key, value in kw.iteritems()]
        argl.extend(kwl)

        # Make a string representation of the args, kw.
        argstr = ', '.join(argl)
        return '%s(%s)'%(func_name, argstr)

    def _is_arbitrary_object(self, object):
        """Return True if the object is an arbitrary non-primitive object.

        As done in appscripting, we assume that if the hex id of the object is
        in its string representation then it is an arbitrary object.
        """
        ob_id = id(object)
        orepr = repr(object)
        hex_id = "%x"%ob_id
        return hex_id.upper() in orepr.upper()

    def _object_as_string(self, object):
        """Return a string representing the object.
        """
        registry = self._registry
        if object in registry:
            # Return script id if the object is known; create the script
            # id on the namespace if needed before that.
            sid = registry.get(object).script_id
            base_id = sid.split('.')[0]
            self.write_script_id_in_namespace(base_id)
            return sid
        else:
            if not self._is_arbitrary_object(object):
                return repr(object)

        # If we get here, we just register the object and call ourselves
        # again to do the needful.
        self.register(object)
        return self._object_as_string(object)

    def _return_as_string(self, object):
        """Return a string given a returned object from a function.
        """
        result = ''
        long_type = long if PY_VER == 2 else int
        ignore = (float, complex, bool, int, long_type, str)
        if object is not None and type(object) not in ignore:
            # If object is not know, register it.
            registry = self._registry
            if object not in registry:
                self.register(object)
            result = registry.get(object).script_id
            # Since this is returned it is known on the namespace.
            known_ids = self._known_ids
            if result not in known_ids:
                known_ids.append(result)
        return result

    def _import_class_string(self, cls):
        """Import a class if needed.
        """
        cname = cls.__name__
        result = ''
        if cname not in __builtin__.__dict__:
            mod = cls.__module__
            typename = '%s.%s'%(mod, cname)
            if typename not in self._known_types:
                result = 'from %s import %s'%(mod, cname)
                self._known_types.append(typename)
        return result



""" A trait type used to access services. """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import TraitType


# Logging.
logger = logging.getLogger(__name__)


class Service(TraitType):
    """ A trait type used to access services.

    Note that this is a trait *type* and hence does *NOT* have traits itself
    (i.e. it does *not* inherit from 'HasTraits').

    """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(
        self, protocol=None, query='', minimize='', maximize='', **metadata
    ):
        """ Constructor. """

        super(Service, self).__init__(**metadata)

        # The protocol that the service must provide.
        self._protocol = protocol

        # The optional query.
        self._query = query

        # The optional name of the trait/property to minimize.
        self._minimize = minimize

        # The optional name of the trait/property to maximize.
        self._maximize = maximize

        return

    ###########################################################################
    # 'TraitType' interface.
    ###########################################################################

    def get(self, obj, trait_name):
        """ Trait type getter. """

        service_registry = self._get_service_registry(obj)


        obj = service_registry.get_service(
            self._protocol, self._query, self._minimize, self._maximize
        )

        return obj

    def set(self, obj, name, value):
        """ Trait type setter. """

        raise SystemError('Service traits cannot be set')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_service_registry(self, obj):
        """ Return the service registry in effect for an object. """

        service_registry = getattr(obj, 'service_registry', None)
        if service_registry is None:
            raise ValueError(
                'The "Service" trait type can only be used within objects ' \
                'that have a reference to a service registry via their ' \
                '"service_registry" trait. \n' \
                'Object %s\nService protocol %s' % (obj, self._protocol)
            )

        return service_registry

#### EOF ######################################################################



""" A collection of command-line tools for building encoded update.xml
files.
"""

class InfoFile:

    update_file = ""
    version = None
    checksum = None

    # A multi-line HTML document describing the changes between
    # this version and the previous version
    description = ""

    @classmethod
    def from_info_file(filename):
        return


def files2xml(filenames):
    """ Given a list of filenames, extracts the app version and log
    information from accompanying files produces an output xml file.

    There are no constraints or restrictions on the names or extensions
    of the input files.  They just need to be accompanied by a sidecar
    file named similarly, but with a ".info" extension, that can be
    loaded by the InfoFile class.
    """
    return






#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   02/04/2005
#
#------------------------------------------------------------------------------

""" Defines the help interface for displaying the help associated with a
    Traits UI View object.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from .toolkit import toolkit

#-------------------------------------------------------------------------
#  Default handler for showing the help associated with a view:
#-------------------------------------------------------------------------


def default_show_help(info, control):
    """ Default handler for showing the help associated with a view.
    """
    toolkit().show_help(info.ui, control)

# The default handler for showing help
show_help = default_show_help

#-------------------------------------------------------------------------
#  Allows an application to change the default show help handler:
#-------------------------------------------------------------------------


def on_help_call(new_show_help=None):
    """ Sets a new global help provider function.

    Parameters
    ----------
    new_show_help : function
        The function to set as the new global help provider

    Returns
    -------
    The previous global help provider function

    Description
    -----------
    The help provider function must have a signature of
    *function*(*info*, *control*), where *info* is a UIInfo object for the
    current view, and *control* is the UI control that invokes the function
    (typically, a **Help** button). It is provided in case the help provider
    needs to position the help window relative to the **Help** button.

    To retrieve the current help provider function, call this function with
    no arguments.
    """
    global show_help

    result = show_help
    if new_show_help is not None:
        show_help = new_show_help
    return result



"""
fs.wrapfs.hidefs
================

Removes resources from a directory listing if they match a given set of wildcards

"""

from fs.wrapfs import WrapFS
from fs.path import iteratepath
from fs.errors import ResourceNotFoundError
import re
import fnmatch


class HideFS(WrapFS):
    """FS wrapper that hides resources if they match a wildcard(s).

    For example, to hide all pyc file and subversion directories from a filesystem::

        hide_fs = HideFS(my_fs, "*.pyc", ".svn")

    """

    def __init__(self, wrapped_fs, *hide_wildcards):
        self._hide_wildcards = [re.compile(fnmatch.translate(wildcard)) for wildcard in hide_wildcards]
        super(HideFS, self).__init__(wrapped_fs)

    def _should_hide(self, path):
        return any(any(wildcard.match(part) for wildcard in self._hide_wildcards)
                for part in iteratepath(path))

    def _encode(self, path):
        if self._should_hide(path):
            raise ResourceNotFoundError(path)
        return path

    def _decode(self, path):
        return path

    def exists(self, path):
        if self._should_hide(path):
            return False
        return super(HideFS, self).exists(path)

    def listdir(self, path="", *args, **kwargs):
        entries = super(HideFS, self).listdir(path, *args, **kwargs)
        entries = [entry for entry in entries if not self._should_hide(entry)]
        return entries

if __name__ == "__main__":
    from fs.osfs import OSFS
    hfs = HideFS(OSFS('~/projects/pyfilesystem'), "*.pyc", ".svn")
    hfs.tree()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the HTML editor factory. HTML editors interpret and display
    HTML-formatted text, but do not modify it.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Str, false

from ..basic_editor_factory import BasicEditorFactory

from ..toolkit import toolkit_object

# Callable that returns the editor to use in the UI.


def html_editor(*args, **traits):
    return toolkit_object('html_editor:SimpleEditor')(*args, **traits)

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Template used to create code blocks embedded in the module comment
block_template = """<center><table width="95%%"><tr><td bgcolor="#ECECEC"><tt>
%s</tt></td></tr></table></center>"""

# Template used to create lists embedded in the module comment
list_template = """<%s>
%s
</%s>"""

#------------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#------------------------------------------------------------------------------


class ToolkitEditorFactory(BasicEditorFactory):
    """ Editor factory for HTML editors.
    """
    #--------------------------------------------------------------------------
    #  Trait definitions:
    #--------------------------------------------------------------------------

    # Should implicit text formatting be converted to HTML?
    format_text = false

    # External objects referenced in the HTML are relative to this url
    base_url = Str

    # The object trait containing the base URL
    base_url_name = Str

    # Should links be opened in an external browser?
    open_externally = false

    #-------------------------------------------------------------------------
    #  Parses the contents of a formatted text string into the corresponding
    #  HTML:
    #-------------------------------------------------------------------------

    def parse_text(self, text):
        """ Parses the contents of a formatted text string into the
            corresponding HTML.
        """
        text = text.replace('\r\n', '\n')
        lines = [('.' + line).strip()[1:] for line in text.split('\n')]
        ind = min(*([self.indent(line) for line in lines
                     if line != ''] + [1000, 1000]))
        if ind >= 1000:
            ind = 0
        lines = [line[ind:] for line in lines]
        new_lines = []
        i = 0
        n = len(lines)
        while i < n:
            line = lines[i]
            m = self.indent(line)
            if m > 0:
                if line[m] in '-*':
                    i, line = self.parse_list(lines, i)
                else:
                    i, line = self.parse_block(lines, i)
                new_lines.append(line)
            else:
                new_lines.append(line)
                i += 1
        text = '\n'.join(new_lines)
        paragraphs = [p.strip() for p in text.split('\n\n')]
        for i, paragraph in enumerate(paragraphs):
            if paragraph[:3].lower() != '<p>':
                paragraphs[i] = '<p>%s</p>' % paragraph
        return '\n'.join(paragraphs)

    #-------------------------------------------------------------------------
    #  Parses a code block:
    #-------------------------------------------------------------------------

    def parse_block(self, lines, i):
        """ Parses a code block.
        """
        m = 1000
        n = len(lines)
        j = i
        while j < n:
            line = lines[j]
            if line != '':
                k = self.indent(line)
                if k == 0:
                    break
                m = min(m, k)
            j += 1
        j -= 1
        while (j > i) and (lines[j] == ''):
            j -= 1
        j += 1
        temp = [(('&nbsp;' * (self.indent(line) - m)) +
                 line.strip()) for line in lines[i: j]]
        return (j, block_template % '\n<br>'.join(temp))

    #-------------------------------------------------------------------------
    #  Parses a list:
    #-------------------------------------------------------------------------

    def parse_list(self, lines, i):
        """ Parses a list.
        """
        line = lines[i]
        m = self.indent(line)
        kind = line[m]
        result = ['<li>' + line[m + 1:].strip()]
        n = len(lines)
        j = i + 1
        while j < n:
            line = lines[j]
            k = self.indent(line)
            if k < m:
                break
            if k == m:
                if line[k] != kind:
                    break
                result.append('<li>' + line[k + 1:].strip())
                j += 1
            elif line[k] in '-*':
                j, line = self.parse_list(lines, j)
                result.append(line)
            else:
                result.append(line.strip())
                j += 1
        style = ['ul', 'ol'][kind == '*']
        return (j, list_template % (style, '\n'.join(result), style))

    #-------------------------------------------------------------------------
    #  Calculates the amount of white space at the beginning of a line:
    #-------------------------------------------------------------------------

    def indent(self, line):
        """ Calculates the amount of white space at the beginning of a line.
        """
        return len(line) - len((line + '.').strip()) + 1

HTMLEditor = ToolkitEditorFactory(klass=html_editor)

#-EOF--------------------------------------------------------------------------



""" Tests for preferences nodes. """


# Standard library imports.
import os, tempfile, unittest
from os.path import join

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from apptools.preferences.api import Preferences
from traits.api import HasTraits, Int, Str


# This module's package.
PKG = 'apptools.preferences.tests'


class PreferencesTestCase(unittest.TestCase):
    """ Tests for preferences nodes. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = Preferences()

        # The filename of the example preferences file.
        self.example = resource_filename(PKG, 'example.ini')

        # A temporary directory that can safely be written to.
        self.tmpdir = tempfile.mkdtemp()


    def tearDown(self):
        """ Called immediately after each test method has been called. """

        # Remove the temporary directory.
        os.removedirs(self.tmpdir)

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_package_global_default_preferences(self):
        """ package global default preferences """

        from apptools.preferences.api import get_default_preferences
        from apptools.preferences.api import set_default_preferences

        set_default_preferences(self.preferences)
        self.assertEqual(self.preferences, get_default_preferences())

        return

    def test_get_and_set_str(self):
        """ get and set str """

        p = self.preferences

        # Set a string preference.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))

        return

    def test_get_and_set_int(self):
        """ get and set int """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set('acme.ui.width', 50)
        self.assertEqual('50', p.get('acme.ui.width'))

        return

    def test_get_and_set_float(self):
        """ get and set float """

        p = self.preferences

        # Note that we can pass an actual 'flaot' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set('acme.ui.ratio', 1.0)
        self.assertEqual('1.0', p.get('acme.ui.ratio'))

        return

    def test_get_and_set_bool(self):
        """ get and set bool """

        p = self.preferences

        # Note that we can pass an actual 'bool' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set('acme.ui.visible', True)
        self.assertEqual('True', p.get('acme.ui.visible'))

        return

    def test_get_and_set_list_of_str(self):
        """ get and set list of str """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set('acme.ui.names', ['fred', 'wilma', 'barney'])
        self.assertEqual("['fred', 'wilma', 'barney']", p.get('acme.ui.names'))

        return

    def test_get_and_set_list_of_int(self):
        """ get and set list of int """

        p = self.preferences

        # Note that we can pass an actual 'int' to 'set', but the preference
        # manager *always* returns preference values as strings.
        p.set('acme.ui.offsets', [1, 2, 3])
        self.assertEqual('[1, 2, 3]', p.get('acme.ui.offsets'))

        return

    def test_empty_path(self):
        """ empty path """

        p = self.preferences

        self.assertRaises(ValueError, p.get, '')
        self.assertRaises(ValueError, p.remove, '')
        self.assertRaises(ValueError, p.set, '', 'a value')

        return

    def test_default_values(self):
        """ default values """

        p = self.preferences

        # Try non-existent names to get the default-default!
        self.assertEqual(None, p.get('bogus'))
        self.assertEqual(None, p.get('acme.bogus'))
        self.assertEqual(None, p.get('acme.ui.bogus'))

        # Try non-existent names to get the specified default.
        self.assertEqual('a value', p.get('bogus', 'a value'))
        self.assertEqual('a value', p.get('acme.bogus', 'a value'))
        self.assertEqual('a value', p.get('acme.ui.bogus', 'a value'))

        return

    def test_keys(self):
        """ keys """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.keys())

        # Set some preferences in the node.
        p.set('a', '1')
        p.set('b', '2')
        p.set('c', '3')

        keys = p.keys()
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        # Set some preferences in a child node.
        p.set('acme.a', '1')
        p.set('acme.b', '2')
        p.set('acme.c', '3')

        keys = p.keys('acme')
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        # And, just to be sure, in a child of the child node ;^)
        p.set('acme.ui.a', '1')
        p.set('acme.ui.b', '2')
        p.set('acme.ui.c', '3')

        keys = p.keys('acme.ui')
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        # Test keys of a non-existent node.
        self.assertEqual([], p.keys('bogus'))
        self.assertEqual([], p.keys('bogus.blargle'))
        self.assertEqual([], p.keys('bogus.blargle.foogle'))

        return

    def test_node(self):
        """ node """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node('acme')
        self.assertNotEqual(None, node)
        self.assertEqual('acme', node.name)
        self.assertEqual('acme', node.path)
        self.assertEqual(p, node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node('acme'))

        # Try a nested path.
        node = p.node('acme.ui')
        self.assertNotEqual(None, node)
        self.assertEqual('ui', node.name)
        self.assertEqual('acme.ui', node.path)
        self.assertEqual(p.node('acme'), node.parent)

        # And just to be sure, a really nested path.
        node = p.node('acme.ui.splash_screen')
        self.assertNotEqual(None, node)
        self.assertEqual('splash_screen', node.name)
        self.assertEqual('acme.ui.splash_screen', node.path)
        self.assertEqual(p.node('acme.ui'), node.parent)

        return

    def test_node_exists(self):
        """ node exists """

        p = self.preferences

        self.assertEqual(True, p.node_exists())
        self.assertEqual(False, p.node_exists('acme'))

        p.node('acme')
        self.assertEqual(True, p.node_exists('acme'))

        return

    def test_node_names(self):
        """ node names """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.node_names())

        # Add some nodes.
        p.node('a')
        p.node('b')
        p.node('c')

        names = p.node_names()
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        # Creatd some nodes in a child node.
        p.node('acme.a')
        p.node('acme.b')
        p.node('acme.c')

        names = p.node_names('acme')
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        # And, just to be sure, in a child of the child node ;^)
        p.node('acme.ui.a')
        p.node('acme.ui.b')
        p.node('acme.ui.c')

        names = p.node_names('acme.ui')
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        # Test keys of a non-existent node.
        self.assertEqual([], p.node_names('bogus'))
        self.assertEqual([], p.node_names('bogus.blargle'))
        self.assertEqual([], p.node_names('bogus.blargle.foogle'))

        return

    def test_clear(self):
        """ clear """

        p = self.preferences

        # Set some values.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))
        p.set('acme.ui.width', 100)
        self.assertEqual('100', p.get('acme.ui.width'))

        # Clear all preferences from the node.
        p.clear('acme.ui')
        self.assertEqual(None, p.get('acme.ui.bgcolor'))
        self.assertEqual(None, p.get('acme.ui.width'))
        self.assertEqual(0, len(p.keys('acme.ui')))

        return

    def test_remove(self):
        """ remove """

        p = self.preferences

        # Set a value.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))

        # Remove it.
        p.remove('acme.ui.bgcolor')
        self.assertEqual(None, p.get('acme.ui.bgcolor'))

        # Make sure we can't remove nodes!
        p.remove('acme.ui')
        self.assertEqual(True, p.node_exists('acme.ui'))

        return

    def test_flush(self):
        """ flush """

        p = self.preferences

        # A temporary .ini file for this test.
        tmp = join(self.tmpdir, 'tmp.ini')

        # This could be set in the constructor of course, its just here we
        # want to use the instance declared in 'setUp'.
        p.filename = tmp

        try:
            # Load the preferences from an 'ini' file.
            p.load(self.example)

            # Flush it.
            p.flush()

            # Load it into a new node.
            p = Preferences()
            p.load(tmp)

            # Make sure it was all loaded!
            self.assertEqual('blue', p.get('acme.ui.bgcolor'))
            self.assertEqual('50', p.get('acme.ui.width'))
            self.assertEqual('1.0', p.get('acme.ui.ratio'))
            self.assertEqual('True', p.get('acme.ui.visible'))
            self.assertEqual('acme ui', p.get('acme.ui.description'))
            self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
            self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
            self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
            self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        finally:
            # Clean up!
            os.remove(tmp)

        return

    def test_load(self):
        """ load """

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))
        self.assertEqual('50', p.get('acme.ui.width'))
        self.assertEqual('1.0', p.get('acme.ui.ratio'))
        self.assertEqual('True', p.get('acme.ui.visible'))
        self.assertEqual('acme ui', p.get('acme.ui.description'))
        self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
        self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
        self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
        self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        return

    def test_load_with_filename_trait_set(self):
        """ load with filename trait set """

        p = self.preferences
        p.filename = self.example

        # Load the preferences from an 'ini' file.
        p.load()

        # Make sure it was all loaded!
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))
        self.assertEqual('50', p.get('acme.ui.width'))
        self.assertEqual('1.0', p.get('acme.ui.ratio'))
        self.assertEqual('True', p.get('acme.ui.visible'))
        self.assertEqual('acme ui', p.get('acme.ui.description'))
        self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
        self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
        self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
        self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))
        self.assertEqual('50', p.get('acme.ui.width'))
        self.assertEqual('1.0', p.get('acme.ui.ratio'))
        self.assertEqual('True', p.get('acme.ui.visible'))
        self.assertEqual('acme ui', p.get('acme.ui.description'))
        self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
        self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
        self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
        self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        return

    def test_save(self):
        """ save """

        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)

        # Make sure it was all loaded!
        self.assertEqual('blue', p.get('acme.ui.bgcolor'))
        self.assertEqual('50', p.get('acme.ui.width'))
        self.assertEqual('1.0', p.get('acme.ui.ratio'))
        self.assertEqual('True', p.get('acme.ui.visible'))
        self.assertEqual('acme ui', p.get('acme.ui.description'))
        self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
        self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
        self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
        self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        # Make a change.
        p.set('acme.ui.bgcolor', 'yellow')

        # Save it to another file.
        tmp = join(self.tmpdir, 'tmp.ini')
        p.save(tmp)

        try:
            # Load it into a new node.
            p = Preferences()
            p.load(tmp)

            # Make sure it was all loaded!
            self.assertEqual('yellow', p.get('acme.ui.bgcolor'))
            self.assertEqual('50', p.get('acme.ui.width'))
            self.assertEqual('1.0', p.get('acme.ui.ratio'))
            self.assertEqual('True', p.get('acme.ui.visible'))
            self.assertEqual('acme ui', p.get('acme.ui.description'))
            self.assertEqual('[1, 2, 3, 4]', p.get('acme.ui.offsets'))
            self.assertEqual("['joe', 'fred', 'jane']", p.get('acme.ui.names'))
            self.assertEqual('splash', p.get('acme.ui.splash_screen.image'))
            self.assertEqual('red', p.get('acme.ui.splash_screen.fgcolor'))

        finally:
            # Clean up!
            os.remove(tmp)

        return

    def SKIPtest_dump(self):
        """ dump """

        # This make look like a weird test, since we don't ever actually check
        # anything, but it is useful for people to see the structure of a
        # preferences hierarchy.
        p = self.preferences

        # Load the preferences from an 'ini' file.
        p.load(self.example)
        p.dump()

        return

    def test_get_inherited(self):
        """ get inherited """

        p = self.preferences

        # Set a string preference.
        p.set('bgcolor', 'red')
        p.set('acme.bgcolor', 'green')
        p.set('acme.ui.bgcolor', 'blue')

        self.assertEqual('blue', p.get('acme.ui.bgcolor', inherit=True))

        # Now remove the 'lowest' layer.
        p.remove('acme.ui.bgcolor')
        self.assertEqual('green', p.get('acme.ui.bgcolor', inherit=True))

        # And the next one.
        p.remove('acme.bgcolor')
        self.assertEqual('red', p.get('acme.ui.bgcolor', inherit=True))

        # And the last one.
        p.remove('bgcolor')
        self.assertEqual(None, p.get('acme.ui.bgcolor', inherit=True))

        return

    def test_add_listener(self):
        """ add listener """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key  = key
            listener.old  = old
            listener.new  = new

            return

        # Add a listener.
        p.add_preferences_listener(listener, 'acme.ui')

        # Set a value and make sure the listener was called.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual(p.node('acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual(None, listener.old)
        self.assertEqual('blue', listener.new)

        # Set it to another value to make sure we get the 'old' value
        # correctly.
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual(p.node('acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual('blue', listener.old)
        self.assertEqual('red', listener.new)

        return

    def test_remove_listener(self):
        """ remove listener """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key  = key
            listener.old  = old
            listener.new  = new

            return

        # Add a listener.
        p.add_preferences_listener(listener, 'acme.ui')

        # Set a value and make sure the listener was called.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual(p.node('acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual(None, listener.old)
        self.assertEqual('blue', listener.new)

        # Remove the listener.
        p.remove_preferences_listener(listener, 'acme.ui')

        # Set a value and make sure the listener was *not* called.
        listener.node = None

        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual(None, listener.node)

        return

    def test_set_with_same_value(self):
        """ set with same value """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key  = key
            listener.old  = old
            listener.new  = new

            return

        # Add a listener.
        p.add_preferences_listener(listener, 'acme.ui')

        # Set a value and make sure the listener was called.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual(p.node('acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual(None, listener.old)
        self.assertEqual('blue', listener.new)

        # Clear out the listener.
        listener.node = None

        # Set the same value and make sure the listener *doesn't* get called.
        p.set('acme.ui.bgcolor', 'blue')
        self.assertEqual(None, listener.node)

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



# for py2/3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import datetime

import pytest
import numpy as np
import netCDF4
from post_gnome import nc_particles

## test the Reader
def test_read_required():
    """
    Does it find the required variables and attributes
    Should be able to set up data_index
    """
    r = nc_particles.Reader('sample.nc')
    assert len(r.times) == 3
    assert np.array_equal(r.data_index, np.array([0, 3, 7, 9]))

def test_read_existing_dataset():
    nc = netCDF4.Dataset('sample.nc')
    r = nc_particles.Reader(nc)
    assert len(r.times) == 3

def test_str():
    r = nc_particles.Reader('sample.nc')
    print(r)
    r.close()
    assert True

## other tests fail (E   RuntimeError: NetCDF: Not a valid ID)
## if this test is here -- no idea why, but I think NetCDF4 isn't cleaning up after itself well
def test_read_variables():
    """
    does it find the data variables ?
    """
    r = nc_particles.Reader('sample.nc')
    # set(), because order doesn't matter
    varnames = set(r.variables)
    assert varnames == set(['latitude', 'depth', 'mass', 'id', 'longitude'])

def test_get_all_timesteps():
    r = nc_particles.Reader('sample.nc')
    data = r.get_all_timesteps(variables=['depth', 'mass', 'id'])
    print(data)
    assert 'depth' in data
    assert 'mass' in data
    assert 'id' in data
    ## better to check actual data, but what can you do?

def test_get_timestep():
    r = nc_particles.Reader('sample.nc')
    data = r.get_timestep(2, variables=['latitude', 'depth', 'mass', 'id', 'longitude'])

    # specific results from the sample file
    assert np.array_equal(data['longitude'], [-88.3, -88.1])
    assert np.array_equal(data['latitude'], [28.1, 28.0])
    assert np.array_equal(data['depth'], [0.0, 0.1])
    assert np.array_equal(data['mass'], [0.05, 0.06])
    assert np.array_equal(data['id'], [1, 3])

def test_get_individual_trajectory():
    r = nc_particles.Reader('sample.nc')

    path = r.get_individual_trajectory(1)
    assert np.array_equal(path['latitude'],  [28.0, 28.05, 28.1])
    assert np.array_equal(path['longitude'], [-88.1, -88.2, -88.3])

def test_get_units():
    r = nc_particles.Reader('sample.nc')

    assert r.get_units('depth') == 'meters'
    assert r.get_units('longitude') == 'degrees_east'

def test_get_attributes():
    r = nc_particles.Reader('sample.nc')

    assert  r.get_attributes('depth') == {'units' : "meters",
                                          'long_name' : "particle depth below sea surface",
                                          'standard_name' : "depth",
                                          'axis' : "z positive down",
                                          }



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the base wxPython EditorFactory class and classes the various
    styles of editors used in a Traits-based user interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import warnings

import wx

from traits.api \
    import TraitError, Any, Bool, Event, Str

from traitsui.editor_factory \
    import EditorFactory as BaseEditorFactory

from editor \
    import Editor

from constants \
    import WindowColor

#-------------------------------------------------------------------------
#  'EditorFactory' class
#   Deprecated alias for traitsui.editor_factory.EditorFactory
#-------------------------------------------------------------------------


class EditorFactory(BaseEditorFactory):
    """ Deprecated alias for traitsui.editor_factory.EditorFactory.
    """

    def __init__(self, *args, **kwds):
        super(EditorFactory, self).__init__(*args, **kwds)
        warnings.warn("DEPRECATED: Use traitsui.editor_factory."
                      ".EditorFactory instead.", DeprecationWarning)

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Base class for simple style editors, which displays a text field
        containing the text representation of the object trait value. Clicking
        in the text field displays an editor-specific dialog box for changing
        the value.
    """

    # Has the left mouse button been pressed:
    left_down = Bool(False)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = self.create_control(parent)
        wx.EVT_LEFT_DOWN(self.control, self._enable_popup_editor)
        wx.EVT_LEFT_UP(self.control, self._show_popup_editor)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Creates the control to use for the simple editor:
    #-------------------------------------------------------------------------

    def create_control(self, parent):
        """ Creates the control to use for the simple editor.
        """
        return wx.TextCtrl(parent, -1, self.str_value, style=wx.TE_READONLY)

    #-------------------------------------------------------------------------
    #  Invokes the pop-up editor for an object trait:
    #
    #  (Normally overridden in a subclass)
    #-------------------------------------------------------------------------

    def popup_editor(self, event):
        """ Invokes the pop-up editor for an object trait.
        """
        pass

    def _enable_popup_editor(self, event):
        """ Mark the left mouse button as being pressed currently.
        """
        self.left_down = True

    def _show_popup_editor(self, event):
        """ Display the popup editor if the left mouse button was pressed
            previously.
        """
        if self.left_down:
            self.left_down = False
            self.popup_editor(event)

#-------------------------------------------------------------------------
#  'TextEditor' class:
#-------------------------------------------------------------------------


class TextEditor(Editor):
    """ Base class for text style editors, which displays an editable text
        field, containing a text representation of the object trait value.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = wx.TextCtrl(parent, -1, self.str_value,
                                   style=wx.TE_PROCESS_ENTER)
        wx.EVT_KILL_FOCUS(self.control, self.update_object)
        wx.EVT_TEXT_ENTER(parent, self.control.GetId(), self.update_object)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Handles the user changing the contents of the edit control:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        try:
            self.value = self.control.GetValue()
        except TraitError as excp:
            pass

#-------------------------------------------------------------------------
#  'ReadonlyEditor' class:
#-------------------------------------------------------------------------


class ReadonlyEditor(Editor):
    """ Base class for read-only style editors, which displays a read-only text
        field, containing a text representation of the object trait value.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # layout_style = 0  # Style for imbedding control in a sizer (override)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        if (self.item.resizable is True) or (self.item.height != -1.0):
            self.control = wx.TextCtrl(
                parent, -1, self.str_value, style=wx.NO_BORDER | wx.TE_MULTILINE | wx.TE_READONLY)
            self.control.SetBackgroundColour(WindowColor)
        else:
            self.control = wx.StaticText(parent, -1, self.str_value,
                                         style=wx.ALIGN_LEFT)
            self.layout_style = 0

        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        new_value = self.str_value
        if (self.item.resizable is True) or (self.item.height != -1.0):
            if self.control.GetValue() != new_value:
                self.control.SetValue(new_value)
        elif self.control.GetLabel() != new_value:
            self.control.SetLabel(new_value)



""" A manager for adapter factories. """


# Enthought library imports.
from traits.api import Dict, HasTraits, Instance, Property

# Local imports.
from abstract_type_system import AbstractTypeSystem
from python_type_system import PythonTypeSystem


class AdapterManager(HasTraits):
    """ A manager for adapter factories. """

    #### 'AdapterManager' interface ###########################################

    # All registered type-scope factories by the type of object that they
    # adapt.
    #
    # The dictionary is keyed by the *name* of the class rather than the class
    # itself to allow for adapter factory proxies to register themselves
    # without having to load and create the factories themselves (i.e., to
    # allow us to lazily load adapter factories contributed by plugins). This
    # is a slight compromise as it is obviously geared towards use in Envisage,
    # but it doesn't affect the API other than allowing a class OR a string to
    # be passed to 'register_adapters'.
    #
    # { String adaptee_class_name : List(AdapterFactory) factories }
    type_factories = Property(Dict)

    # All registered instance-scope factories by the object that they adapt.
    #
    # { id(obj) : List(AdapterFactory) factories }
    instance_factories = Property(Dict)

    # The type system used by the manager (it determines 'is_a' relationships
    # and type MROs etc). By default we use standard Python semantics.
    type_system = Instance(AbstractTypeSystem, PythonTypeSystem())

    #### Private interface ####################################################

    # All registered type-scope factories by the type of object that they
    # adapt.
    _type_factories = Dict

    # All registered instance-scope factories by the object that they adapt.
    _instance_factories = Dict

    ###########################################################################
    # 'AdapterManager' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_type_factories(self):
        """ Returns all registered type-scope factories. """

        return self._type_factories.copy()

    def _get_instance_factories(self):
        """ Returns all registered instance-scope factories. """

        return self._instance_factories.copy()

    #### Methods ##############################################################

    def adapt(self, adaptee, target_class, *args, **kw):
        """ Returns an adapter that adapts an object to the target class.

        'adaptee' is the object that we want to adapt.
        'target_class' is the class that the adaptee should be adapted to.

        Returns None if no such adapter exists.

        """

        # If the object is already an instance of the target class then we
        # simply return it.
        if self.type_system.is_a(adaptee, target_class):
            adapter = adaptee

        # Otherwise, look at each class in the adaptee's MRO to see if there
        # is an adapter factory registered that can adapt the object to the
        # target class.
        else:
            # Look for instance-scope adapters first.
            adapter = self._adapt_instance(adaptee, target_class, *args, **kw)

            # If no instance-scope adapter was found then try type-scope
            # adapters.
            if adapter is None:
                for adaptee_class in self.type_system.get_mro(type(adaptee)):
                    adapter = self._adapt_type(
                        adaptee, adaptee_class, target_class, *args, **kw
                    )
                    if adapter is not None:
                        break

        return adapter

    def register_instance_adapters(self, factory, obj):
        """ Registers an instance-scope adapter factory.

        A factory can be in exactly one manager (as it uses the manager's type
        system).

        """

        factories = self._instance_factories.setdefault(id(obj), [])
        factories.append(factory)

        # A factory can be in exactly one manager.
        factory.adapter_manager = self

        return

    def unregister_instance_adapters(self, factory, obj):
        """ Unregisters an instance scope adapter factory.

        A factory can be in exactly one manager (as it uses the manager's type
        system).

        """

        factories = self._instance_factories.setdefault(id(obj), [])
        if factory in factories:
            factories.remove(factory)

        # A factory can be in exactly one manager.
        factory.adapter_manager = None

        return

    def register_type_adapters(self, factory, adaptee_class):
        """ Registers a type-scope adapter factory.

        'adaptee_class' can be either a class object or the name of a class.

        A factory can be in exactly one manager (as it uses the manager's type
        system).

        """

        if isinstance(adaptee_class, basestring):
            adaptee_class_name = adaptee_class

        else:
            adaptee_class_name = self._get_class_name(adaptee_class)

        factories = self._type_factories.setdefault(adaptee_class_name, [])
        factories.append(factory)

        # A factory can be in exactly one manager.
        factory.adapter_manager = self

        return

    def unregister_type_adapters(self, factory):
        """ Unregisters a type-scope adapter factory. """

        for adaptee_class_name, factories in self._type_factories.items():
            if factory in factories:
                factories.remove(factory)

        # The factory is no longer deemed to be part of this manager.
        factory.adapter_manager = None

        return

    #### DEPRECATED ###########################################################

    def register_adapters(self, factory, adaptee_class):
        """ Registers an adapter factory.

        'adaptee_class' can be either a class object or the name of a class.

        A factory can be in exactly one manager (as it uses the manager's type
        system).

        """

        print 'DEPRECATED: use "register_type_adapters" instead.'

        self.register_type_adapters(factory, adaptee_class)

        return

    def unregister_adapters(self, factory):
        """ Unregisters an adapter factory. """

        print 'DEPRECATED: use "unregister_type_adapters" instead.'

        self.unregister_type_adapters(factory)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _adapt_instance(self, adaptee, target_class, *args, **kw):
        """ Returns an adapter that adaptes an object to the target class.

        Returns None if no such adapter exists.

        """

        for factory in self._instance_factories.get(id(adaptee), []):
            adapter = factory.adapt(adaptee, target_class, *args, **kw)
            if adapter is not None:
                break

        else:
            adapter = None

        return adapter

    def _adapt_type(self, adaptee, adaptee_class, target_class, *args, **kw):
        """ Returns an adapter that adapts an object to the target class.

        Returns None if no such adapter exists.

        """

        adaptee_class_name = self._get_class_name(adaptee_class)

        for factory in self._type_factories.get(adaptee_class_name, []):
            adapter = factory.adapt(adaptee, target_class, *args, **kw)
            if adapter is not None:
                break

        else:
            adapter = None

        return adapter

    def _get_class_name(self, klass):
        """ Returns the full class name for a class. """

        return "%s.%s" % (klass.__module__, klass.__name__)

#### EOF ######################################################################



""" Tests for dynamic notifiers with `dispatch='new'`.

Dynamic notifiers created with the `dispatch='new'` option dispatch event
notifications on a new thread. The class handling the dispatch,
`NewTraitChangeNotifyWrapper`, is a subclass of `TraitChangeNotifyWrapper`.
Most of the functionality of the class is thus already covered by the
`TestDynamicNotifiers` test case, and we only need to test that the
notification really occurs on a separate thread.

"""
import thread
import time

from traits.api import Float, HasTraits
from traits.testing.unittest_tools import unittest


class Foo(HasTraits):
    foo = Float


class TestNewNotifiers(unittest.TestCase):
    """ Tests for dynamic notifiers with `dispatch='new'`. """

    def test_notification_on_separate_thread(self):
        notifications = []

        def on_foo_notifications(obj, name, old, new):
            thread_id = thread.get_ident()
            event = (thread_id, obj, name, old, new)
            notifications.append(event)

        obj = Foo()
        obj.on_trait_change(on_foo_notifications, 'foo', dispatch='new')

        obj.foo = 3
        # Wait for a while to make sure the notification has finished.
        time.sleep(0.1)

        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0][1:], (obj, 'foo', 0, 3))

        this_thread_id = thread.get_ident()
        self.assertNotEqual(this_thread_id, notifications[0][0])


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/22/2004
#
#------------------------------------------------------------------------------

""" Trait definition for a wxPython-based font.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Trait, TraitHandler, TraitError

#-------------------------------------------------------------------------
#  Convert a string into a valid 'wxFont' object (if possible):
#-------------------------------------------------------------------------

# Mapping of strings to valid wxFont families
font_families = {
    'default': wx.DEFAULT,
    'decorative': wx.DECORATIVE,
    'roman': wx.ROMAN,
    'script': wx.SCRIPT,
    'swiss': wx.SWISS,
    'modern': wx.MODERN
}

# Mapping of strings to wxFont styles
font_styles = {
    'slant': wx.SLANT,
    'italic': wx.ITALIC
}

# Mapping of strings wxFont weights
font_weights = {
    'light': wx.LIGHT,
    'bold': wx.BOLD
}

# Strings to ignore in text representations of fonts
font_noise = ['pt', 'point', 'family']

#-------------------------------------------------------------------------
#  Converts a wx.Font into a string description of itself:
#-------------------------------------------------------------------------


def font_to_str(font):
    """ Converts a wx.Font into a string description of itself.
    """
    weight = {wx.LIGHT: ' Light',
              wx.BOLD: ' Bold'}.get(font.GetWeight(), '')
    style = {wx.SLANT: ' Slant',
             wx.ITALIC: ' Italic'}.get(font.GetStyle(), '')
    underline = ''
    if font.GetUnderlined():
        underline = ' underline'
    return '%s point %s%s%s%s' % (
           font.GetPointSize(), font.GetFaceName(), style, weight, underline)

#-------------------------------------------------------------------------
#  Create a TraitFont object from a string description:
#-------------------------------------------------------------------------


def create_traitsfont(value):
    """ Create a TraitFont object from a string description.
    """
    if isinstance(value, wx.Font):
        value = font_to_str(value)

    point_size = None
    family = wx.DEFAULT
    style = wx.NORMAL
    weight = wx.NORMAL
    underline = 0
    facename = []
    for word in value.split():
        lword = word.lower()
        if lword in font_families:
            family = font_families[lword]
        elif lword in font_styles:
            style = font_styles[lword]
        elif lword in font_weights:
            weight = font_weights[lword]
        elif lword == 'underline':
            underline = 1
        elif lword not in font_noise:
            if point_size is None:
                try:
                    point_size = int(lword)
                    continue
                except:
                    pass
            facename.append(word)
    return TraitsFont(point_size or 10, family, style, weight, underline,
                      ' '.join(facename))

#-------------------------------------------------------------------------
#  'TraitsFont' class:
#-------------------------------------------------------------------------


class TraitsFont(wx.Font):
    """ A Traits-specific wx.Font.
    """
    #-------------------------------------------------------------------------
    #  Returns the pickleable form of a TraitsFont object:
    #-------------------------------------------------------------------------

    def __reduce_ex__(self, protocol):
        """ Returns the pickleable form of a TraitsFont object.
        """
        return (create_traitsfont, (font_to_str(self), ))

    #-------------------------------------------------------------------------
    #  Returns a printable form of the font:
    #-------------------------------------------------------------------------

    def __str__(self):
        """ Returns a printable form of the font.
        """
        return font_to_str(self)

#-------------------------------------------------------------------------
#  'TraitWXFont' class'
#-------------------------------------------------------------------------


class TraitWXFont(TraitHandler):
    """ Ensures that values assigned to a trait attribute are valid font
    descriptor strings; the value actually assigned is the corresponding
    TraitsFont.
    """
    #-------------------------------------------------------------------------
    #  Validates that the value is a valid font:
    #-------------------------------------------------------------------------

    def validate(self, object, name, value):
        """ Validates that the value is a valid font descriptor string. If so,
        it returns the corresponding TraitsFont; otherwise, it raises a
        TraitError.
        """
        if value is None:
            return None

        try:
            return create_traitsfont(value)
        except:
            pass

        raise TraitError(object, name, 'a font descriptor string',
                         repr(value))

    def info(self):
        return ("a string describing a font (e.g. '12 pt bold italic "
                "swiss family Arial' or 'default 12')")

#-------------------------------------------------------------------------
#  Define a wxPython specific font trait:
#-------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the FontEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Font trait which lead to this file getting
# imported. This leads to a circular import when declaring a Font trait.


def get_font_editor(*args, **traits):
    from font_editor import ToolkitEditorFactory
    return ToolkitEditorFactory(*args, **traits)

fh = TraitWXFont()
WxFont = Trait(
    wx.SystemSettings_GetFont(
        wx.SYS_DEFAULT_GUI_FONT),
    fh,
    editor=get_font_editor)



"""
fs.opener
=========

Open filesystems via a URI.

There are occasions when you want to specify a filesystem from the command line
or in a config file. This module enables that functionality, and can return an
FS object given a filesystem specification in a URI-like syntax (inspired by
the syntax of http://commons.apache.org/vfs/filesystems.html).

The `OpenerRegistry` class maps the protocol (file, ftp etc.) on to an Opener
object, which returns an appropriate filesystem object and path.  You can
create a custom opener registry that opens just the filesystems you require, or
use the opener registry defined here (also called `opener`) that can open any
supported filesystem.

The `parse` method of an `OpenerRegsitry` object returns a tuple of an FS
object a path. Here's an example of how to use the default opener registry::

    >>> from fs.opener import opener
    >>> opener.parse('ftp://ftp.mozilla.org/pub')
    (<fs.ftpfs.FTPFS object at 0x96e66ec>, u'pub')

You can use use the `opendir` method, which just returns an FS object. In the
example above, `opendir` will return a FS object for the directory `pub`::

    >>> opener.opendir('ftp://ftp.mozilla.org/pub')
    <SubFS: <FTPFS ftp.mozilla.org>/pub>

If you are just interested in a single file, use the `open` method of a registry
which returns a file-like object, and has the same signature as FS objects and
the `open` builtin::

    >>> opener.open('ftp://ftp.mozilla.org/pub/README')
    <fs.ftpfs._FTPFile object at 0x973764c>

The `opendir` and `open` methods can also be imported from the top-level of
this module for sake of convenience.  To avoid shadowing the builtin `open`
method, they are named `fsopendir` and `fsopen`. Here's how you might import
them::

    from fs.opener import fsopendir, fsopen


"""

__all__ = ['OpenerError',
           'NoOpenerError',
           'OpenerRegistry',
           'opener',
           'fsopen',
           'fsopendir',
           'OpenerRegistry',
           'Opener',
           'OSFSOpener',
           'ZipOpener',
           'RPCOpener',
           'FTPOpener',
           'SFTPOpener',
           'MemOpener',
           'DebugOpener',
           'TempOpener',
           'S3Opener',
           'TahoeOpener',
           'DavOpener',
           'HTTPOpener']

from fs.path import pathsplit, join, iswildcard, normpath
from fs.osfs import OSFS
from fs.filelike import FileWrapper
from os import getcwd
import os.path
import re
from urlparse import urlparse

class OpenerError(Exception):
    """The base exception thrown by openers"""
    pass

class NoOpenerError(OpenerError):
    """Thrown when there is no opener for the given protocol"""
    pass

def _expand_syspath(path):
    if path is None:
        return path
    if path.startswith('\\\\?\\') and not path.startswith('\\\\?\\UNC\\'):
        path = path[4:]
    path = os.path.expanduser(os.path.expandvars(path))
    path = os.path.normpath(os.path.abspath(path))
    return path

def _parse_credentials(url):
    scheme = None
    if '://' in url:
        scheme, url = url.split('://', 1)
    username = None
    password = None
    if '@' in url:
        credentials, url = url.split('@', 1)
        if ':' in credentials:
            username, password = credentials.split(':', 1)
        else:
            username = credentials
    if scheme is not None:
        url = '%s://%s' % (scheme, url)
    return username, password, url

def _parse_name(fs_name):
    if '#' in fs_name:
        fs_name, fs_name_params = fs_name.split('#', 1)
        return fs_name, fs_name_params
    else:
        return fs_name, None

def _split_url_path(url):
    if '://' not in url:
        url = 'http://' + url
    scheme, netloc, path, _params, _query, _fragment = urlparse(url)
    url = '%s://%s' % (scheme, netloc)
    return url, path


def _FSClosingFile(fs, file_object, mode):
    original_close = file_object.close

    def close():
        try:
            fs.close()
        except:
            pass
        return original_close()
    file_object.close = close
    return file_object


class OpenerRegistry(object):

    """An opener registry that  stores a number of opener objects used to parse FS URIs"""

    re_fs_url = re.compile(r'''
^
(.*?)
:\/\/

(?:
(?:(.*?)@(.*?))
|(.*?)
)

(?:
!(.*?)$
)*$
''', re.VERBOSE)


    def __init__(self, openers=[]):
        self.registry = {}
        self.openers = {}
        self.default_opener = 'osfs'
        for opener in openers:
            self.add(opener)

    @classmethod
    def split_segments(self, fs_url):
        match = self.re_fs_url.match(fs_url)
        return match

    def get_opener(self, name):
        """Retrieve an opener for the given protocol

        :param name: name of the opener to open
        :raises NoOpenerError: if no opener has been registered of that name

        """
        if name not in self.registry:
            raise NoOpenerError("No opener for %s" % name)
        index = self.registry[name]
        return self.openers[index]

    def add(self, opener):
        """Adds an opener to the registry

        :param opener: a class derived from fs.opener.Opener

        """

        index = len(self.openers)
        self.openers[index] = opener
        for name in opener.names:
            self.registry[name] = index

    def parse(self, fs_url, default_fs_name=None, writeable=False, create_dir=False, cache_hint=True):
        """Parses a FS url and returns an fs object a path within that FS object
        (if indicated in the path). A tuple of (<FS instance>, <path>) is returned.

        :param fs_url: an FS url
        :param default_fs_name: the default FS to use if none is indicated (defaults is OSFS)
        :param writeable: if True, a writeable FS will be returned
        :param create_dir: if True, then the directory in the FS will be created

        """

        orig_url = fs_url
        match = self.split_segments(fs_url)

        if match:
            fs_name, credentials, url1, url2, path = match.groups()
            if credentials:
                fs_url = '%s@%s' % (credentials, url1)
            else:
                fs_url = url2
            path = path or ''
            fs_url = fs_url or ''
            if ':' in fs_name:
                fs_name, sub_protocol = fs_name.split(':', 1)
                fs_url = '%s://%s' % (sub_protocol, fs_url)
            if '!' in path:
                paths = path.split('!')
                path = paths.pop()
                fs_url = '%s!%s' % (fs_url, '!'.join(paths))

            fs_name = fs_name or self.default_opener
        else:
            fs_name = default_fs_name or self.default_opener
            fs_url = _expand_syspath(fs_url)
            path = ''

        fs_name,  fs_name_params = _parse_name(fs_name)
        opener = self.get_opener(fs_name)

        if fs_url is None:
            raise OpenerError("Unable to parse '%s'" % orig_url)

        fs, fs_path = opener.get_fs(self, fs_name, fs_name_params, fs_url, writeable, create_dir)
        fs.cache_hint(cache_hint)

        if fs_path and iswildcard(fs_path):
            pathname, resourcename = pathsplit(fs_path or '')
            if pathname:
                fs = fs.opendir(pathname)
            return fs, resourcename

        fs_path = join(fs_path, path)

        if create_dir and fs_path:
            if not fs.getmeta('read_only', False):
                fs.makedir(fs_path, allow_recreate=True)

        pathname, resourcename = pathsplit(fs_path or '')
        if pathname and resourcename:
            fs = fs.opendir(pathname)
            fs_path = resourcename

        return fs, fs_path or ''

    def open(self, fs_url, mode='r', **kwargs):
        """Opens a file from a given FS url

        If you intend to do a lot of file manipulation, it would likely be more
        efficient to do it directly through the an FS instance (from `parse` or
        `opendir`). This method is fine for one-offs though.

        :param fs_url: a FS URL, e.g. ftp://ftp.mozilla.org/README
        :param mode: mode to open file file
        :rtype: a file

        """

        writeable = 'w' in mode or 'a' in mode or '+' in mode
        fs, path = self.parse(fs_url, writeable=writeable)
        file_object = fs.open(path, mode)

        file_object = _FSClosingFile(fs, file_object, mode)
        #file_object.fs = fs
        return file_object

    def getcontents(self, fs_url, mode='rb', encoding=None, errors=None, newline=None):
        """Gets the contents from a given FS url (if it references a file)

        :param fs_url: a FS URL e.g. ftp://ftp.mozilla.org/README

        """
        fs, path = self.parse(fs_url)
        return fs.getcontents(path, mode, encoding=encoding, errors=errors, newline=newline)

    def opendir(self, fs_url, writeable=True, create_dir=False):
        """Opens an FS object from an FS URL

        :param fs_url: an FS URL e.g. ftp://ftp.mozilla.org
        :param writeable: set to True (the default) if the FS must be writeable
        :param create_dir: create the directory references by the FS URL, if
            it doesn't already exist

        """
        fs, path = self.parse(fs_url, writeable=writeable, create_dir=create_dir)
        if path and '://' not in fs_url:
            # A shortcut to return an OSFS rather than a SubFS for os paths
            return OSFS(fs_url)
        if path:
            fs = fs.opendir(path)
        return fs


class Opener(object):
    """The base class for openers

    Opener follow a very simple protocol. To create an opener, derive a class
    from `Opener` and define a classmethod called `get_fs`, which should have the following signature::

        @classmethod
        def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):

    The parameters of `get_fs` are as follows:

     * `fs_name` the name of the opener, as extracted from the protocol part of the url,
     * `fs_name_params` reserved for future use
     * `fs_path` the path part of the url
     * `writeable` if True, then `get_fs` must return an FS that can be written to
     * `create_dir` if True then `get_fs` should attempt to silently create the directory references in path

    In addition to `get_fs` an opener class should contain
    two class attributes: names and desc. `names` is a list of protocols that
    list opener will opener. `desc` is an English description of the individual opener syntax.

    """
    pass


class OSFSOpener(Opener):
    names = ['osfs', 'file']
    desc = """OS filesystem opener, works with any valid system path. This is the default opener and will be used if you don't indicate which opener to use.

    examples:
    * file://relative/foo/bar/baz.txt (opens a relative file)
    * file:///home/user (opens a directory from a absolute path)
    * osfs://~/ (open the user's home directory)
    * foo/bar.baz (file:// is the default opener)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.osfs import OSFS

        path = os.path.normpath(fs_path)
        if create_dir and not os.path.exists(path):
            from fs.osfs import _os_makedirs
            _os_makedirs(path)
        dirname, resourcename = os.path.split(fs_path)
        osfs = OSFS(dirname)
        return osfs, resourcename

class ZipOpener(Opener):
    names = ['zip', 'zip64']
    desc = """Opens zip files. Use zip64 for > 2 gigabyte zip files, if you have a 64 bit processor.

    examples:
    * zip://myzip.zip (open a local zip file)
    * zip://myzip.zip!foo/bar/insidezip.txt (reference a file insize myzip.zip)
    * zip:ftp://ftp.example.org/myzip.zip (open a zip file stored on a ftp server)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):

        zip_fs, zip_path = registry.parse(fs_path)
        if zip_path is None:
            raise OpenerError('File required for zip opener')
        if zip_fs.exists(zip_path):
            if writeable:
                open_mode = 'r+b'
            else:
                open_mode = 'rb'
        else:
            open_mode = 'w+'
        if zip_fs.hassyspath(zip_path):
            zip_file = zip_fs.getsyspath(zip_path)
        else:
            zip_file = zip_fs.open(zip_path, mode=open_mode)

        _username, _password, fs_path = _parse_credentials(fs_path)

        from fs.zipfs import ZipFS
        if zip_file is None:
            zip_file = fs_path

        mode = 'r'
        if writeable:
            mode = 'a'

        allow_zip_64 = fs_name.endswith('64')

        zipfs = ZipFS(zip_file, mode=mode, allow_zip_64=allow_zip_64)
        return zipfs, None


class RPCOpener(Opener):
    names = ['rpc']
    desc = """An opener for filesystems server over RPC (see the fsserve command).

examples:
rpc://127.0.0.1:8000 (opens a RPC server running on local host, port 80)
rpc://www.example.org (opens an RPC server on www.example.org, default port 80)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.rpcfs import RPCFS
        _username, _password, fs_path = _parse_credentials(fs_path)
        if '://' not in fs_path:
            fs_path = 'http://' + fs_path

        scheme, netloc, path, _params, _query, _fragment = urlparse(fs_path)

        rpcfs = RPCFS('%s://%s' % (scheme, netloc))

        if create_dir and path:
            rpcfs.makedir(path, recursive=True, allow_recreate=True)

        return rpcfs, path or None


class FTPOpener(Opener):
    names = ['ftp']
    desc = """An opener for FTP (File Transfer Protocl) server

examples:
* ftp://ftp.mozilla.org (opens the root of ftp.mozilla.org)
* ftp://ftp.example.org/foo/bar (opens /foo/bar on ftp.mozilla.org)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.ftpfs import FTPFS
        username, password, fs_path = _parse_credentials(fs_path)

        scheme, _netloc, _path, _params, _query, _fragment = urlparse(fs_path)
        if not scheme:
            fs_path = 'ftp://' + fs_path
        scheme, netloc, path, _params, _query, _fragment = urlparse(fs_path)

        dirpath, resourcepath = pathsplit(path)
        url = netloc

        ftpfs = FTPFS(url, user=username or '', passwd=password or '', follow_symlinks=(fs_name_params == "symlinks"))
        ftpfs.cache_hint(True)

        if create_dir and path:
            ftpfs.makedir(path, recursive=True, allow_recreate=True)

        if dirpath:
            ftpfs = ftpfs.opendir(dirpath)

        if not resourcepath:
            return ftpfs, None
        else:
            return ftpfs, resourcepath


class SFTPOpener(Opener):
    names = ['sftp']
    desc = """An opener for SFTP (Secure File Transfer Protocol) servers

examples:
* sftp://username:password@example.org (opens sftp server example.org with username and password
* sftp://example.org (opens example.org with public key authentication)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        username, password, fs_path = _parse_credentials(fs_path)

        from fs.sftpfs import SFTPFS

        credentials = {}
        if username is not None:
            credentials['username'] = username
        if password is not None:
            credentials['password'] = password

        if '/' in fs_path:
            addr, fs_path = fs_path.split('/', 1)
        else:
            addr = fs_path
            fs_path = '/'

        fs_path, resourcename = pathsplit(fs_path)

        host = addr
        port = None
        if ':' in host:
            addr, port = host.rsplit(':', 1)
            try:
                port = int(port)
            except ValueError:
                pass
            else:
                host = (addr, port)

        if create_dir:
            sftpfs = SFTPFS(host, root_path='/', **credentials)
            if not sftpfs._transport.is_authenticated():
                sftpfs.close()
                raise OpenerError('SFTP requires authentication')
            sftpfs = sftpfs.makeopendir(fs_path)
            return sftpfs, None

        sftpfs = SFTPFS(host, root_path=fs_path, **credentials)
        if not sftpfs._transport.is_authenticated():
            sftpfs.close()
            raise OpenerError('SFTP requires authentication')

        return sftpfs, resourcename


class MemOpener(Opener):
    names = ['mem', 'ram']
    desc = """Creates an in-memory filesystem (very fast but contents will disappear on exit).
Useful for creating a fast temporary filesystem for serving or mounting with fsserve or fsmount.
NB: If you user fscp or fsmv to copy/move files here, you are effectively deleting them!

examples:
* mem:// (opens a new memory filesystem)
* mem://foo/bar (opens a new memory filesystem with subdirectory /foo/bar)    """

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        from fs.memoryfs import MemoryFS
        memfs = MemoryFS()
        if create_dir:
            memfs = memfs.makeopendir(fs_path)
        return memfs, None


class DebugOpener(Opener):
    names = ['debug']
    desc = """For developers -- adds debugging information to output.

example:
    * debug:ftp://ftp.mozilla.org (displays details of calls made to a ftp filesystem)"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        from fs.wrapfs.debugfs import DebugFS
        if fs_path:
            fs, _path = registry.parse(fs_path, writeable=writeable, create_dir=create_dir)
            return DebugFS(fs, verbose=False), None
        if fs_name_params == 'ram':
            from fs.memoryfs import MemoryFS
            return DebugFS(MemoryFS(), identifier=fs_name_params, verbose=False), None
        else:
            from fs.tempfs import TempFS
            return DebugFS(TempFS(), identifier=fs_name_params, verbose=False), None


class TempOpener(Opener):
    names = ['temp']
    desc = """Creates a temporary filesystem that is erased on exit.
Probably only useful for mounting or serving.
NB: If you use fscp or fsmv to copy/move files here, you are effectively deleting them!

example:
* temp://"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        from fs.tempfs import TempFS
        from fs.wrapfs.lazyfs import LazyFS
        fs = LazyFS((TempFS,(),{"identifier":fs_name_params}))
        return fs, fs_path


class S3Opener(Opener):
    names = ['s3']
    desc = """Opens a filesystem stored on Amazon S3 storage
    The environment variables AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY should be set"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.s3fs import S3FS

        username, password, bucket = _parse_credentials(fs_path)
        path = ''
        if '/' in bucket:
            bucket, path = fs_path.split('/', 1)

        fs = S3FS(bucket,
                  aws_access_key=username or None,
                  aws_secret_key=password or None)

        if path:
            dirpath, resourcepath = pathsplit(path)
            if dirpath:
                fs = fs.opendir(dirpath)
            path = resourcepath

        return fs, path


class TahoeOpener(Opener):
    names = ['tahoe']
    desc = """Opens a Tahoe-LAFS filesystem

    example:
    * tahoe://http://pubgrid.tahoe-lafs.org/uri/URI:DIR2:h5bkxelehowscijdb [...]"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.contrib.tahoelafs import TahoeLAFS

        if '/uri/' not in fs_path:
            raise OpenerError("""Tahoe-LAFS url should be in the form <url>/uri/<dicap>""")

        url, dircap = fs_path.split('/uri/')
        path = ''
        if '/' in dircap:
            dircap, path = dircap.split('/', 1)

        fs = TahoeLAFS(dircap, webapi=url)

        if '/' in path:
            dirname, _resourcename = pathsplit(path)
            if create_dir:
                fs = fs.makeopendir(dirname)
            else:
                fs = fs.opendir(dirname)
            path = ''

        return fs, path


class DavOpener(Opener):
    names = ['dav']
    desc = """Opens a WebDAV server

example:
* dav://example.org/dav"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.contrib.davfs import DAVFS

        url = fs_path

        if '://' not in url:
            url = 'http://' + url

        scheme, url = url.split('://', 1)

        username, password, url = _parse_credentials(url)

        credentials = None
        if username or password:
            credentials = {}
            if username:
                credentials['username'] = username
            if password:
                credentials['password'] = password

        url = '%s://%s' % (scheme, url)

        fs = DAVFS(url, credentials=credentials)

        return fs, ''

class HTTPOpener(Opener):
    names = ['http', 'https']
    desc = """HTTP file opener. HTTP only supports reading files, and not much else.

example:
* http://www.example.org/index.html"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        from fs.httpfs import HTTPFS
        if '/' in fs_path:
            dirname, resourcename = fs_path.rsplit('/', 1)
        else:
            dirname = fs_path
            resourcename = ''
        fs = HTTPFS('http://' + dirname)
        return fs, resourcename

class UserDataOpener(Opener):
    names = ['appuserdata', 'appuser']
    desc = """Opens a filesystem for a per-user application directory.

The 'domain' should be in the form <author name>:<application name>.<version> (the author name and version are optional).

example:
* appuserdata://myapplication
* appuserdata://examplesoft:myapplication
* appuserdata://anotherapp.1.1
* appuserdata://examplesoft:anotherapp.1.3"""

    FSClass = 'UserDataFS'

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):
        import fs.appdirfs
        fs_class = getattr(fs.appdirfs, cls.FSClass)
        if ':' in fs_path:
            appauthor, appname = fs_path.split(':', 1)
        else:
            appauthor = None
            appname = fs_path

        if '/' in appname:
            appname, path = appname.split('/', 1)
        else:
            path = ''

        if '.' in appname:
            appname, appversion = appname.split('.', 1)
        else:
            appversion = None

        fs = fs_class(appname, appauthor=appauthor, version=appversion, create=create_dir)

        if '/' in path:
            subdir, path = path.rsplit('/', 1)
            if create_dir:
                fs = fs.makeopendir(subdir, recursive=True)
            else:
                fs = fs.opendir(subdir)

        return fs, path

class SiteDataOpener(UserDataOpener):
    names = ['appsitedata', 'appsite']

    desc = """Opens a filesystem for an application site data directory.

The 'domain' should be in the form <author name>:<application name>.<version> (the author name and version are optional).

example:
* appsitedata://myapplication
* appsitedata://examplesoft:myapplication
* appsitedata://anotherapp.1.1
* appsitedata://examplesoft:anotherapp.1.3"""

    FSClass = 'SiteDataFS'

class UserCacheOpener(UserDataOpener):
    names = ['appusercache', 'appcache']

    desc = """Opens a filesystem for an per-user application cache directory.

The 'domain' should be in the form <author name>:<application name>.<version> (the author name and version are optional).

example:
* appusercache://myapplication
* appusercache://examplesoft:myapplication
* appusercache://anotherapp.1.1
* appusercache://examplesoft:anotherapp.1.3"""

    FSClass = 'UserCacheFS'


class UserLogOpener(UserDataOpener):
    names = ['appuserlog', 'applog']

    desc = """Opens a filesystem for an application site data directory.

The 'domain' should be in the form <author name>:<application name>.<version> (the author name and version are optional).

example:
* appuserlog://myapplication
* appuserlog://examplesoft:myapplication
* appuserlog://anotherapp.1.1
* appuserlog://examplesoft:anotherapp.1.3"""

    FSClass = 'UserLogFS'


class MountOpener(Opener):
    names = ['mount']
    desc = """Mounts other filesystems on a 'virtual' filesystem

The path portion of the FS URL should be a path to an ini file, where the keys are the mount point, and the values are FS URLs to mount.

The following is an example of such an ini file:

    [fs]
    resources=appuser://myapp/resources
    foo=~/foo
    foo/bar=mem://

    [fs2]
    bar=~/bar

example:
* mount://fs.ini
* mount://fs.ini!resources
* mount://fs.ini:fs2"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):

        from fs.mountfs import MountFS
        from ConfigParser import ConfigParser
        cfg = ConfigParser()

        if '#' in fs_path:
            path, section = fs_path.split('#', 1)
        else:
            path = fs_path
            section = 'fs'

        cfg.readfp(registry.open(path))

        mount_fs = MountFS()
        for mount_point, mount_path in cfg.items(section):
            mount_fs.mount(mount_point, registry.opendir(mount_path, create_dir=create_dir))
        return mount_fs, ''


class MultiOpener(Opener):
    names = ['multi']
    desc = """Combines other filesystems in to a single filesystem.

The path portion of the FS URL should be a path to an ini file, where the keys are the mount point, and the values are FS URLs to mount.

The following is an example of such an ini file:

    [templates]
    dir1=templates/foo
    dir2=templates/bar

example:
* multi://fs.ini"""

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path, writeable, create_dir):

        from fs.multifs import MultiFS
        from ConfigParser import ConfigParser
        cfg = ConfigParser()

        if '#' in fs_path:
            path, section = fs_path.split('#', 1)
        else:
            path = fs_path
            section = 'fs'

        cfg.readfp(registry.open(path))

        multi_fs = MultiFS()
        for name, fs_url in cfg.items(section):
            multi_fs.addfs(name, registry.opendir(fs_url, create_dir=create_dir))
        return multi_fs, ''


opener = OpenerRegistry([OSFSOpener,
                         ZipOpener,
                         RPCOpener,
                         FTPOpener,
                         SFTPOpener,
                         MemOpener,
                         DebugOpener,
                         TempOpener,
                         S3Opener,
                         TahoeOpener,
                         DavOpener,
                         HTTPOpener,
                         UserDataOpener,
                         SiteDataOpener,
                         UserCacheOpener,
                         UserLogOpener,
                         MountOpener,
                         MultiOpener
                         ])

fsopen = opener.open
fsopendir = opener.opendir




from fs.multifs import MultiFS
from fs.memoryfs import MemoryFS
import unittest

from six import b

class TestMultiFS(unittest.TestCase):
    
    def test_auto_close(self):
        """Test MultiFS auto close is working"""       
        multi_fs = MultiFS()
        m1 = MemoryFS()
        m2 = MemoryFS()
        multi_fs.addfs('m1', m1)
        multi_fs.addfs('m2', m2)
        self.assert_(not m1.closed)
        self.assert_(not m2.closed)
        multi_fs.close()
        self.assert_(m1.closed)
        self.assert_(m2.closed)
        
    def test_no_auto_close(self):
        """Test MultiFS auto close can be disables"""
        multi_fs = MultiFS(auto_close=False)
        m1 = MemoryFS()
        m2 = MemoryFS()
        multi_fs.addfs('m1', m1)
        multi_fs.addfs('m2', m2)
        self.assert_(not m1.closed)
        self.assert_(not m2.closed)
        multi_fs.close()
        self.assert_(not m1.closed)
        self.assert_(not m2.closed)
    
    
    def test_priority(self):
        """Test priority order is working"""
        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setcontents("name", b("m1"))
        m2.setcontents("name", b("m2"))
        m3.setcontents("name", b("m3"))
        multi_fs = MultiFS(auto_close=False)
        multi_fs.addfs("m1", m1)
        multi_fs.addfs("m2", m2)
        multi_fs.addfs("m3", m3)
        self.assert_(multi_fs.getcontents("name") == b("m3"))
        
        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setcontents("name", b("m1"))
        m2.setcontents("name", b("m2"))
        m3.setcontents("name", b("m3"))
        multi_fs = MultiFS(auto_close=False)
        multi_fs.addfs("m1", m1)
        multi_fs.addfs("m2", m2, priority=10)
        multi_fs.addfs("m3", m3)
        self.assert_(multi_fs.getcontents("name") == b("m2"))        
        
        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setcontents("name", b("m1"))
        m2.setcontents("name", b("m2"))
        m3.setcontents("name", b("m3"))
        multi_fs = MultiFS(auto_close=False)
        multi_fs.addfs("m1", m1)
        multi_fs.addfs("m2", m2, priority=10)
        multi_fs.addfs("m3", m3, priority=10)
        self.assert_(multi_fs.getcontents("name") == b("m3"))
        
        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setcontents("name", b("m1"))
        m2.setcontents("name", b("m2"))
        m3.setcontents("name", b("m3"))
        multi_fs = MultiFS(auto_close=False)
        multi_fs.addfs("m1", m1, priority=11)
        multi_fs.addfs("m2", m2, priority=10)
        multi_fs.addfs("m3", m3, priority=10)
        self.assert_(multi_fs.getcontents("name") == b("m1"))
        



""" (Pyface) Action for displaying a help doc.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!


# Standard library imports.
import logging
from os.path import isabs, join, normpath
from subprocess import Popen
import sys

# Enthought library imports.
from envisage.api import IExtensionPointUser, IExtensionRegistry
from pyface.workbench.action.workbench_action import WorkbenchAction
from traits.api import Instance, provides, Property
from pyface.api import ImageResource

from apptools.help.help_plugin.api import HelpDoc

# Local imports
from util import get_sys_prefix_relative_filename

# Logging.
logger = logging.getLogger(__name__)

# This module's parent package.
PARENT = '.'.join(__name__.split('.')[:-2])

# Implementation of the ImageResource class to be used for the DocAction class.
@provides(IExtensionPointUser)
class DocImageResource(ImageResource):
    """ Implementation of the ImageResource class to be used for the DocAction
    class.
    Overrides the '_image_not_found' trait in the base ImageResource class.
    """

    _image_not_found = ImageResource('document')

class DocAction(WorkbenchAction):
    """ (Pyface) Action for displaying a help doc.
    """


    ### Action interface ##############################################

    # Image associated with this action instance.
    image = Property

    ### IExtensionPointUser interface

    extension_registry = Property(Instance(IExtensionRegistry))

    def _get_extension_registry(self):
        return self.window.application.extension_registry

    def _get_image(self):
        """ Returns the image to be used for this DocAction instance.
        """
        # The current implementation searches for an image file matching
        # 'name' in all of the image paths. If such a file is not to be found,
        # the '_image_not_found' file for the DocImageResourceClass is used.
        return DocImageResource(self.name)

    ### HelpDocAction interface

    # Help doc associated with this action.
    my_help_doc = Instance(HelpDoc)

    def _my_help_doc_default(self):
        exns = self.extension_registry.get_extensions(PARENT + '.help_docs')
        for hd in exns:
            if hd.label == self.name:
                return hd
        return None

    def _get_filename(self, doc):
        filename = None
        if doc is not None:
            if doc.url:
                filename = doc.filename
            else:
                filename = get_sys_prefix_relative_filename(doc.filename)
        return filename

    def perform(self, event):
        """ Perform the action by displaying the document.
        """

        filename = self._get_filename(self.my_help_doc)
        if filename is not None:
            if self.my_help_doc.url or self.my_help_doc.viewer == 'browser':
                import webbrowser
                try:
                    webbrowser.open(filename)
                except (OSError, webbrowser.Error), msg:
                    logger.error('Could not open page in browser for '+ \
                        'Document "%s":\n\n' % self.my_help_doc.label + \
                        str(msg) + '\n\nTry changing Dcoument Preferences.')
            elif self.my_help_doc.viewer is not None:
                # Run the viewer, passing it the filename
                try:
                    Popen([self.my_help_doc.viewer, filename])
                except OSError, msg:
                    logger.error('Could not execute program for Document' + \
                        ' "%s":\n\n ' % self.my_help_doc.label + str(msg) + \
                        '\n\nTry changing Document Preferences.')



from .i_action_set import IActionSet
from .i_action_manager_builder import IActionManagerBuilder

from .abstract_action_manager_builder import AbstractActionManagerBuilder
from .action import Action
from .action_set import ActionSet
from .group import Group
from .menu import Menu
from .tool_bar import ToolBar




#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   05/20/2005
#
#-------------------------------------------------------------------------

""" Defines KeyBinding and KeyBindings classes, which manage the mapping of
    keystroke events into method calls on controller objects that are supplied
    by the application.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    Any,
    Event,
    HasPrivateTraits,
    HasStrictTraits,
    Instance,
    List,
    Property,
    Str,
    cached_property,
    on_trait_change)

from .api import HGroup, Item, KeyBindingEditor, ListEditor, View, toolkit

from traits.trait_base import SequenceTypes

#-------------------------------------------------------------------------
#  Key binding trait definition:
#-------------------------------------------------------------------------

# Trait definition for key bindings
Binding = Str(event='binding', editor=KeyBindingEditor())

#-------------------------------------------------------------------------
#  'KeyBinding' class:
#-------------------------------------------------------------------------


class KeyBinding(HasStrictTraits):
    """ Binds one or two keystrokes to a method.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # First key binding
    binding1 = Binding

    # Second key binding
    binding2 = Binding

    # Description of what application function the method performs
    description = Str

    # Name of controller method the key is bound to
    method_name = Str

    # KeyBindings object that "owns" the KeyBinding
    owner = Instance('KeyBindings')

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View(
        HGroup(
            Item('binding1'),
            Item('binding2'),
            Item('description', style='readonly'),
            show_labels=False
        )
    )

    #-------------------------------------------------------------------------
    #  Handles a binding trait being changed:
    #-------------------------------------------------------------------------

    def _binding_changed(self):
        if self.owner is not None:
            self.owner.binding_modified = self

#-------------------------------------------------------------------------
#  'KeyBindings' class:
#-------------------------------------------------------------------------


class KeyBindings(HasPrivateTraits):
    """ A set of key bindings.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Set of defined key bindings (redefined dynamically)
    bindings = List(KeyBinding)

    # Optional prefix to add to each method name
    prefix = Str

    # Optional suffix to add to each method name
    suffix = Str

    #-- Private Traits -------------------------------------------------------

    # The (optional) list of controllers associated with this KeyBindings
    # object. The controllers may also be provided with the 'do' method:
    controllers = List(transient=True)

    # The 'parent' KeyBindings object of this one (if any):
    parent = Instance('KeyBindings', transient=True)

    # The root of the KeyBindings tree this object is part of:
    root = Property(depends_on='parent')

    # The child KeyBindings of this object (if any):
    children = List(transient=True)

    # Event fired when one of the contained KeyBinding objects is changed
    binding_modified = Event(KeyBinding)

    # Control that currently has the focus (if any)
    focus_owner = Any(transient=True)

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View([Item('bindings',
                             style='custom',
                             show_label=False,
                             editor=ListEditor(style='custom')),
                        '|{Click on an entry field, then press the key to '
                        'assign. Double-click a field to clear it.}<>'],
                       title='Update Key Bindings',
                       kind='livemodal',
                       resizable=True,
                       width=0.4,
                       height=0.4)

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, *bindings, **traits):
        super(KeyBindings, self).__init__(**traits)

        if (len(bindings) == 1) and isinstance(bindings[0], SequenceTypes):
            bindings = bindings[0]

        n = len(bindings)
        self.add_trait('bindings', List(KeyBinding, minlen=n,
                                        maxlen=n,
                                        mode='list'))
        self.bindings = [binding.set(owner=self) for binding in bindings]

    #-------------------------------------------------------------------------
    #  Processes a keyboard event:
    #-------------------------------------------------------------------------

    def do(self, event, controllers=[], *args, **kw):
        """ Processes a keyboard event.
        """
        if isinstance(controllers, dict):
            controllers = controllers.values()
        elif not isinstance(controllers, SequenceTypes):
            controllers = [controllers]
        else:
            controllers = list(controllers)

        return self._do(toolkit().key_event_to_name(event),
                        controllers, args, kw.get('recursive', False))

    #-------------------------------------------------------------------------
    #  Merges another set of key bindings into this set:
    #-------------------------------------------------------------------------

    def merge(self, key_bindings):
        """ Merges another set of key bindings into this set.
        """
        binding_dic = {}
        for binding in self.bindings:
            binding_dic[binding.method_name] = binding

        for binding in key_bindings.bindings:
            binding2 = binding_dic.get(binding.method_name)
            if binding2 is not None:
                binding2.binding1 = binding.binding1
                binding2.binding2 = binding.binding2

    #-------------------------------------------------------------------------
    #  Returns a clone of the KeyBindings object:
    #-------------------------------------------------------------------------

    def clone(self, **traits):
        """ Returns a clone of the KeyBindings object.
        """
        return self.__class__(*self.bindings, **traits).set(
            **self.get('prefix', 'suffix'))

    #-------------------------------------------------------------------------
    #  Dispose of the object:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Dispose of the object.
        """
        if self.parent is not None:
            self.parent.children.remove(self)

        del self.controllers
        del self.children
        del self.bindings

        self.parent = self._root = self.focus_owner = None

    #-------------------------------------------------------------------------
    #  Edits a possibly hierarchical set of KeyBindings:
    #-------------------------------------------------------------------------

    def edit(self):
        """ Edits a possibly hierarchical set of KeyBindings.
        """
        bindings = list(set(self.root._get_bindings([])))
        bindings.sort(key=lambda x: '%s%02d' % (x.binding1[-1:], x.binding1))
        KeyBindings(bindings).edit_traits()

    #-------------------------------------------------------------------------
    #  Returns the current binding for a specified key (if any):
    #-------------------------------------------------------------------------

    def key_binding_for(self, binding, key_name):
        """ Returns the current binding for a specified key (if any).
        """
        if key_name != '':
            for a_binding in self.bindings:
                if ((a_binding is not binding) and
                    ((key_name == a_binding.binding1) or
                     (key_name == a_binding.binding2))):
                    return a_binding

        return None

    #-- Property Implementations ---------------------------------------------

    @cached_property
    def _get_root(self):
        root = self
        while root.parent is not None:
            root = root.parent

        return root

    #-- Event Handlers -------------------------------------------------------

    def _binding_modified_changed(self, binding):
        """ Handles a binding being changed.
        """
        binding1 = binding.binding1
        binding2 = binding.binding2
        for a_binding in self.bindings:
            if binding is not a_binding:
                if binding1 == a_binding.binding1:
                    a_binding.binding1 = ''
                if binding1 == a_binding.binding2:
                    a_binding.binding2 = ''
                if binding2 == a_binding.binding1:
                    a_binding.binding1 = ''
                if binding2 == a_binding.binding2:
                    a_binding.binding2 = ''

    def _focus_owner_changed(self, old, new):
        """ Handles the focus owner being changed.
        """
        if old is not None:
            old.border_size = 0

    @on_trait_change('children[]')
    def _children_modified(self, removed, added):
        """ Handles child KeyBindings being added to the object.
        """
        for item in added:
            item.parent = self

    #-- object Method Overrides ----------------------------------------------

    #-------------------------------------------------------------------------
    #  Restores the state of a previously pickled object:
    #-------------------------------------------------------------------------

    def __setstate__(self, state):
        """ Restores the state of a previously pickled object.
        """
        n = len(state['bindings'])
        self.add_trait('bindings', List(KeyBinding, minlen=n, maxlen=n))
        self.__dict__.update(state)
        self.bindings = self.bindings[:]

    #-- Private Methods ------------------------------------------------------

    def _get_bindings(self, bindings):
        """ Returns all of the bindings of this object and all of its children.
        """
        bindings.extend(self.bindings)
        for child in self.children:
            child._get_bindings(bindings)

        return bindings

    def _do(self, key_name, controllers, args, recursive):
        """ Process the specified key for the specified set of controllers for
            this KeyBindings object and all of its children.
        """
        # Search through our own bindings for a match:
        for binding in self.bindings:
            if (key_name == binding.binding1) or (
                    key_name == binding.binding2):
                method_name = '%s%s%s' % (
                              self.prefix, binding.method_name, self.suffix)
                for controller in (controllers + self.controllers):
                    method = getattr(controller, method_name, None)
                    if method is not None:
                        result = method(*args)
                        if result is not False:
                            return True

                if binding.method_name == 'edit_bindings':
                    self.edit()
                    return True

        # If recursive, continue searching through a children's bindings:
        if recursive:
            for child in self.children:
                if child._do(key_name, controllers, args, recursive):
                    return True

        # Indicate no one processed the key:
        return False



"""
fs.wrapfs.limitsizefs
=====================

An FS wrapper class for limiting the size of the underlying FS.

This module provides the class LimitSizeFS, an FS wrapper that can limit the
total size of files stored in the wrapped FS.

"""

from __future__ import with_statement

from fs.errors import *
from fs.path import *
from fs.base import FS, threading, synchronize
from fs.wrapfs import WrapFS
from fs.filelike import FileWrapper


class LimitSizeFS(WrapFS):
    """FS wrapper class to limit total size of files stored."""

    def __init__(self, fs, max_size):
        super(LimitSizeFS,self).__init__(fs)
        if max_size < 0:
            try:
                max_size = fs.getmeta("total_space") + max_size
            except NoMetaError:
                msg = "FS doesn't report total_size; "\
                      "can't use negative max_size"
                raise ValueError(msg)
        self.max_size = max_size
        self._size_lock = threading.Lock()
        self._file_sizes = PathMap()
        self.cur_size = self._get_cur_size()

    def __getstate__(self):
        state = super(LimitSizeFS,self).__getstate__()
        del state["cur_size"]
        del state["_size_lock"]
        del state["_file_sizes"]
        return state

    def __setstate__(self, state):
        super(LimitSizeFS,self).__setstate__(state)
        self._size_lock = threading.Lock()
        self._file_sizes = PathMap()
        self.cur_size = self._get_cur_size()

    def _get_cur_size(self,path="/"):
        return sum(self.getsize(f) for f in self.walkfiles(path))

    def getsyspath(self, path, allow_none=False):
        #  If people could grab syspaths, they could route around our
        #  size protection; no dice!
        if not allow_none:
            raise NoSysPathError(path)
        return None

    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        path = relpath(normpath(path))
        with self._size_lock:
            try:
                size = self.getsize(path)
            except ResourceNotFoundError:
                size = 0
            f = super(LimitSizeFS,self).open(path,
                                             mode=mode,
                                             buffering=buffering,
                                             errors=errors,
                                             newline=newline,
                                             line_buffering=line_buffering,
                                             **kwargs)
            if "w" not in mode:
                self._set_file_size(path,None,1)
            else:
                self.cur_size -= size
                size = 0
                self._set_file_size(path,0,1)
            return LimitSizeFile(f,mode,size,self,path)

    def _set_file_size(self,path,size,incrcount=None):
        try:
            (cursize,count) = self._file_sizes[path]
        except KeyError:
            count = 0
            try:
                cursize = self.getsize(path)
            except ResourceNotFoundError:
                cursize = 0
        if size is None:
            size = cursize
        if count is not None:
            count += 1
        if count == 0:
            del self._file_sizes[path]
        else:
            self._file_sizes[path] = (size,count)

    def setcontents(self, path, data, chunk_size=64*1024):
        f = None
        try:
            f = self.open(path, 'wb')
            if hasattr(data, 'read'):
                chunk = data.read(chunk_size)
                while chunk:
                    f.write(chunk)
                    chunk = data.read(chunk_size)
            else:
                f.write(data)
        finally:
            if f is not None:
                f.close()

    def _file_closed(self, path):
        self._set_file_size(path,None,-1)

    def _ensure_file_size(self, path, size, shrink=False):
        with self._size_lock:
            try:
                (cur_size,_) = self._file_sizes[path]
            except KeyError:
                try:
                    cur_size = self.getsize(path)
                except ResourceNotFoundError:
                    cur_size = 0
                self._set_file_size(path,cur_size,1)
            diff = size - cur_size
            if diff > 0:
                if self.cur_size + diff > self.max_size:
                    raise StorageSpaceError("write")
                self.cur_size += diff
                self._set_file_size(path,size)
                return size
            elif diff < 0 and shrink:
                self.cur_size += diff
                self._set_file_size(path,size)
                return size
            else:
                return cur_size

    #  We force use of several base FS methods,
    #  since they will fall back to writing out each file
    #  and thus will route through our size checking logic.
    def copy(self, src, dst, **kwds):
        FS.copy(self,src,dst,**kwds)

    def copydir(self, src, dst, **kwds):
        FS.copydir(self,src,dst,**kwds)

    def move(self, src, dst, **kwds):
        if self.getmeta("atomic.rename",False):
            if kwds.get("overwrite",False) or not self.exists(dst):
                try:
                    self.rename(src,dst)
                    return
                except FSError:
                    pass
        FS.move(self, src, dst, **kwds)

    def movedir(self, src, dst, **kwds):
        overwrite = kwds.get("overwrite",False)
        if self.getmeta("atomic.rename",False):
            if kwds.get("overwrite",False) or not self.exists(dst):
                try:
                    self.rename(src,dst)
                    return
                except FSError:
                    pass
        FS.movedir(self,src,dst,**kwds)

    def rename(self, src, dst):
        if self.getmeta("atomic.rename",False):
            try:
                dst_size = self._get_cur_size(dst)
            except ResourceNotFoundError:
                dst_size = 0
            super(LimitSizeFS,self).rename(src,dst)
            with self._size_lock:
                self.cur_size -= dst_size
                self._file_sizes.pop(src,None)
        else:
            if self.isdir(src):
                self.movedir(src,dst)
            else:
                self.move(src,dst)

    def remove(self, path):
        with self._size_lock:
            try:
                (size,_) = self._file_sizes[path]
            except KeyError:
                size = self.getsize(path)
            super(LimitSizeFS,self).remove(path)
            self.cur_size -= size
            self._file_sizes.pop(path,None)

    def removedir(self, path, recursive=False, force=False):
        #  Walk and remove directories by hand, so they we
        #  keep the size accounting precisely up to date.
        for nm in self.listdir(path):
            if not force:
                raise DirectoryNotEmptyError(path)
            cpath = pathjoin(path,nm)
            try:
                if self.isdir(cpath):
                    self.removedir(cpath,force=True)
                else:
                    self.remove(cpath)
            except ResourceNotFoundError:
                pass
        super(LimitSizeFS,self).removedir(path,recursive=recursive)

    def getinfo(self, path):
        info = super(LimitSizeFS,self).getinfo(path)
        try:
            info["size"] = max(self._file_sizes[path][0],info["size"])
        except KeyError:
            pass
        return info

    def getsize(self, path):
        size = super(LimitSizeFS,self).getsize(path)
        try:
            size = max(self._file_sizes[path][0],size)
        except KeyError:
            pass
        return size



class LimitSizeFile(FileWrapper):
    """Filelike wrapper class for use by LimitSizeFS."""

    def __init__(self, file, mode, size, fs, path):
        super(LimitSizeFile,self).__init__(file,mode)
        self.size = size
        self.fs = fs
        self.path = path
        self._lock = fs._lock

    @synchronize
    def _write(self, data, flushing=False):
        pos = self.wrapped_file.tell()
        new_size = self.fs._ensure_file_size(self.path, pos+len(data))
        res = super(LimitSizeFile,self)._write(data, flushing)
        self.size = new_size
        return res

    @synchronize
    def _truncate(self, size):
        new_size = self.fs._ensure_file_size(self.path,size,shrink=True)
        res = super(LimitSizeFile,self)._truncate(size)
        self.size = new_size
        return res

    @synchronize
    def close(self):
        super(LimitSizeFile,self).close()
        self.fs._file_closed(self.path)





#-----------------------------------------------------------------------------
# Name:        springtabs.py
# Purpose:     Tab-bar control that pops up windows when clicked
#
# Author:      Rob McMullen
#
# Created:     2008
# RCS-ID:      $Id: $
# Copyright:   (c) 2007 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
"""SpringTabs

This module provides popup windows from a group of tabs

"""

import os, sys, struct, Queue, threading, time, socket
from cStringIO import StringIO

import wx
import wx.stc
from wx.lib.buttons import GenToggleButton

import logging
log = logging.getLogger(__name__)


class FakePopupWindow(wx.MiniFrame):
    def __init__(self, parent, style=None):
        super(FakePopupWindow, self).__init__(parent, style = wx.NO_BORDER |wx.FRAME_FLOAT_ON_PARENT | wx.FRAME_NO_TASKBAR)
        #self.Bind(wx.EVT_KEY_DOWN , self.OnKeyDown)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)

    def fix_sizer(self, child):
        pass

    def OnChar(self, evt):
        #print("OnChar: keycode=%s" % evt.GetKeyCode())
        self.GetParent().GetEventHandler().ProcessEvent(evt)

    def Position(self, position, size):
        #print("pos=%s size=%s" % (position, size))
        self.Move((position[0]+size[0], position[1]+size[1]))

    def SetPosition(self, position):
        #print("pos=%s" % (position))
        self.Move((position[0], position[1]))

    def ActivateParent(self):
        """Activate the parent window
        @postcondition: parent window is raised

        """
        parent = self.GetParent()
        parent.Raise()
        parent.SetFocus()

    def OnFocus(self, evt):
        """Raise and reset the focus to the parent window whenever
        we get focus.
        @param evt: event that called this handler

        """
        #log.debug("OnFocus: set focus to %s" % str(self.GetParent()))
        self.ActivateParent()
        evt.Skip()


#class RealPopupWindow(wx.PopupWindow):
class RealPopupWindow(wx.PopupTransientWindow):
    def fix_sizer(self, child):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(child, 1, wx.EXPAND)
        self.SetSizer(sizer)

    def OnDismiss(self):
        print "DISMISSED!!!!!"
        wx.CallAfter(self.GetParent().clear_popup)


class SpringTabItemRenderer(object):
    def __init__(self):
        self.spacing_between_items = 8

    def OnPaint(self, item, evt):
        (width, height) = item.GetClientSizeTuple()
        x1 = y1 = 0
        x2 = width-1
        y2 = height-1

        dc = wx.PaintDC(item)
        if item.hover:
            self.DrawHoverBackground(item, dc)
        else:
            brush = item.GetBackgroundBrush(dc)
            if brush is not None:
                dc.SetBackground(brush)
                dc.Clear()

        item.DrawLabel(dc, width, height)
        self.DrawHoverDecorations(item, dc, width, height)

        #log.debug("button %s: pressed=%s" % (item.GetLabel(), not item.up))

    def DrawHoverBackground(self, item, dc):
        brush = wx.Brush(item.faceDnClr, wx.SOLID)
        dc.SetBackground(brush)
        dc.Clear()

    def DrawHoverDecorations(self, item, dc, width, height):
        pass

    def draw_notification(self, dc, x, y, w, h, num, springtabs):
        if num > 99:
            s = "99+"
        else:
            s = str(num)
        tw, th = dc.GetTextExtent(s)
        dc.SetBrush(springtabs.notification_brush)
        dc.SetPen(springtabs.notification_pen)
        dc.SetTextBackground(springtabs.notification_background)
        dc.SetTextForeground(springtabs.notification_text)
        if tw > w - h/2:
            w2 = tw + h/2
            x -= (w2 - w) / 2
            w = w2
        dc.DrawRoundedRectangle(x, y, w, h, h / 2)
        dc.DrawText(s, x + w/2 - tw/2, y + h/2 - th/2)


class SpringTabItemHorizontalRenderer(SpringTabItemRenderer):
    default_direction = "down"

    def __init__(self, popup_direction="default"):
        SpringTabItemRenderer.__init__(self)
        if popup_direction == "default":
            popup_direction = self.default_direction

        if popup_direction == "down":
            self.popdown = True
        elif popup_direction == "up":
            self.popdown = False
        else:
            raise TypeError("popup_direction %s not valid for horizontal renderer" % popup_direction)


class SpringTabItemVerticalRenderer(SpringTabItemRenderer):
    default_direction = "right"

    def __init__(self, popup_direction="default"):
        SpringTabItemRenderer.__init__(self)
        if popup_direction == "default":
            popup_direction = self.default_direction

        if popup_direction == "left":
            self.popleft = True
        elif popup_direction == "right":
            self.popleft = False
        else:
            raise TypeError("popup_direction %s not valid for vertical renderer" % popup_direction)

    def DoGetBestSize(self, item):
        """
        Overridden base class virtual.  Determines the best size of the
        button based on the label and bezel size.
        """
        h, w, useMin = item._GetLabelSize()
        width = w + 2 * item.border + item.bezelWidth - 1
        height = h + 2 * item.border + item.bezelWidth - 1 + self.spacing_between_items
        #log.debug("width=%d height=%d" % (width, height))
        return (width, height)

    def DrawLabel(self, item, dc, width, height, dx=0, dy=0):
        dc.SetFont(item.GetFont())
        if item.IsEnabled():
            dc.SetTextForeground(item.GetForegroundColour())
        else:
            dc.SetTextForeground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT))
        label = item.GetLabel()
        th, tw = dc.GetTextExtent(label)
        #dc.DrawText(label, (width-tw)/2+dx, (height-th)/2+dy)
        if self.popleft:
            x = (width-tw)/2+dx + tw
            y = dy + item.border + self.spacing_between_items/2
            dc.DrawRotatedText(label, x, y, 270.0)
        else:
            x = (width-tw)/2+dx
            y = height + dy - item.border - self.spacing_between_items/2
            dc.DrawRotatedText(label, x, y, 90.0)
        num = item.notification_count
        if num > 0:
            parent = item.GetParent()
            dc.SetFont(parent.notification_font)
            tw, th = dc.GetTextExtent("0123456789")
            x = item.border
            w = width - 2*item.border
            h = th + 2*item.border
            if self.popleft:
                y = height - h - item.border
            else:
                y = item.border
            self.draw_notification(dc, x, y, w, h, num, parent)

    def set_popup_width(self, parent, popup, child):
        pw, ph = popup.GetSizeTuple()
        pcw, pch = popup.GetClientSizeTuple()
        try:
            cw, ch = child.DoGetBestSize()
        except AttributeError:
            cw, ch = child.GetSizeTuple()
        log.debug("popup size=%s  popup client size=%s  child=%s" % (str((pw, ph)), str((pcw, pch)), str((cw, ch))))

        # The client size may be smaller than the popup window if the popup
        # has a border decoration.
        diffwidth =  pw - pcw
        diffheight =  ph - pch

        # The popup will be at least as tall as the SpringTabs panel
        width, height = parent.GetSizeTuple()
        if ph < height:
            ph = height
        pw = min(cw + diffwidth, parent.max_popup_width)
        popup.SetSize(wx.Size(pw, ph))
        return width, height, pw, ph

    def showPopup(self, parent, item, show=True):
        popup, child = item.getPopup()
        if show:
            if hasattr(child, 'activateSpringTab'):
                child.activateSpringTab()

            # Calculate the child's width twice because the popup itself may
            # not have a valid height the first time through if the popup has
            # never been shown before, and the popup height will be needed by
            # the child to calculate if a scrollbar is needed.
            self.set_popup_width(parent, popup, child)
            width, height, pw, ph = self.set_popup_width(parent, popup, child)

            x, y = parent.ClientToScreenXY(width, 0)
            if self.popleft:
                x -= width + pw
            #log.debug("popping up at %s" % str((x,y)))
            child.SetPosition(wx.Point(0, 0))
            popup.SetPosition(wx.Point(x, y))
            wx.CallAfter(item.setPopupFocusCallback)
            popup.Popup()
        else:
            popup.Dismiss()


class SpringTabItem(GenToggleButton):
    def __init__(self, parent, id=-1, label='', window_cb=None, available_cb=None, **kwargs):
        """Creates a springtab button
        
        This button on the springtab is linked to a callback function that
        creates a window on demand when the button is clicked.
        
        @param parent: parent window: a L{SpringTab} instance
        
        @param id: optional id of the new button
        
        @param label: label of the springtab button
        
        @param window_cb: callback to create the contents of the popup.  This
        functor will be passed two arguments: the first is the popup window
        determined from L{SpringTab.getNewPopup} to be used as the parent
        window for the popup contents.  The second argument is the kwargs dict.
        
        @param available_cb: callback to see if the button should be displayed
        in the L{SpringTab}.  This callback takes a single argument; the
        kwargs dict and should return a boolean indicating whether or not the
        button will be visible.
        
        @param kwargs: dict containing any keyword arguments necessary to
        create the popup contents or check for button availability.
        """
        self.border = 2
        self.hover = False

        GenToggleButton.__init__(self, parent, id, label)
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnEnter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeave)

        self.window_cb = window_cb
        self.available_cb = available_cb
        if self.available_cb:
            self.Show(self.available_cb(**kwargs))
        self.kwargs = kwargs
        self.popup = None
        self.notification_count = 0

    def InitColours(self):
        faceClr = self.GetBackgroundColour()
        r, g, b = faceClr.Get()
        fr, fg, fb = max(0,r-32), max(0,g-32), max(0,b-32)
        #log.debug(str((fr, fg, fb)))
        self.faceDnClr = wx.Colour(fr, fg, fb)
        sr, sg, sb = max(0,r-32), max(0,g-32), max(0,b-32)
        self.shadowPen = wx.Pen(wx.Colour(sr,sg,sb), 1, wx.SOLID)
        hr, hg, hb = min(255,r+64), min(255,g+64), min(255,b+64)
        self.highlightPen = wx.Pen(wx.Colour(hr,hg,hb), 1, wx.SOLID)
        self.focusClr = wx.Colour(hr, hg, hb)

    def DoGetBestSize(self):
        return self.GetParent().getRenderer().DoGetBestSize(self)

    def DrawLabel(self, dc, width, height, dx=0, dy=0):
        self.GetParent().getRenderer().DrawLabel(self, dc, width, height, dx, dy)

    def OnPaint(self, evt):
        self.GetParent().getRenderer().OnPaint(self, evt)

    def SetToggle(self, flag, check_popup=True):
        self.up = not flag
        print "SETTOGGLE! up=", self.up
        if self.up:
            # with PopupTransientWindow, need to force not hovering so it
            # doesn't take a second click to process
            self.hover = False
        if check_popup:
            self.GetParent().setRadio(self)
        self.Refresh()

    def OnLeftDown(self, event):
        if not self.IsEnabled():
            return
        self.saveUp = self.up
        self.up = not self.up
        self.GetParent().setRadio(self)
        self.SetFocus()
        self.Refresh()

    def OnEnter(self, evt):
        self.hover = True
        self.Refresh()

    def OnLeave(self, evt):
        self.hover = False
        self.Refresh()

    def getPopup(self):
        parent = self.GetParent()
        if self.popup is None:
            self.popup = parent.getNewPopup()
            self.popup.Bind(wx.EVT_ACTIVATE, self.OnActivate)

            # Create the window using the popup as the parent
            self.window_cb(self.popup, parent.task, **self.kwargs)
            windowlist = self.popup.GetChildren()
            log.debug("Creating popup: %s" % windowlist)
            if len(windowlist) == 0:
                raise RuntimeError("Popup window creation failed!")
            child = windowlist[0]
#            child.Bind(wx.EVT_SET_FOCUS, self.OnChildFocus)
            child.Bind(wx.EVT_KILL_FOCUS, self.OnLoseChildFocus)
            child.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)
            #child.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
            #child.Bind(wx.EVT_CHAR, self.on_char)
            self.popup.fix_sizer(child)
        windowlist = self.popup.GetChildren()
        if len(windowlist) == 0:
            raise RuntimeError("Popup window creation failed!")
        child = windowlist[0]
        return self.popup, child

    def on_char(self, evt):
        print("OnChar: keycode=%s, popup=%s" % (evt.GetKeyCode(), evt.GetEventObject()))
        popup = evt.GetEventObject()
        print popup.GetChildren()
        #popup.GetParent().GetEventHandler().ProcessEvent(evt)
        evt.Skip()

    def on_key_down(self, evt):
        print("on_key_down: keycode=%s, popup=%s" % (evt.GetKeyCode(), evt.GetEventObject()))
        popup = evt.GetEventObject()
        print popup.GetChildren()
        popup.GetParent().GetEventHandler().ProcessEvent(evt)
        evt.Skip()

    def on_char_hook(self, evt):
        """ESC handler that closes the popup

        """
        print("on_char_hook: keycode=%s, popup=%s" % (evt.GetKeyCode(), evt.GetEventObject()))
        if evt.GetKeyCode() == wx.WXK_ESCAPE:
            wx.CallAfter(self.GetParent().cancel_popup)
        else:
            evt.Skip()

    def OnActivate(self, evt):
        #log.debug("Activating %s: %s, shown=%s" % (self.GetLabel(), evt.GetActive(), self.popup.IsShown()))
        if self.popup.IsShown():
            evt.Skip()

    def setPopupFocusCallback(self):
        """Callback for use within wx.CallAfter to prevent focus being set
        after the control has been removed.
        """
        popup, child = self.getPopup()
        if popup.IsShown():
            #log.debug("setting focus to %s" % self.GetLabel())
            child.SetFocus()

    # FIXME: This attempt, using OnChildFocus, setPopupsLoseFocusCallback, and
    # OnChildLoseFocus, was to see if I could find out when the focus was set
    # to something other than one of the popup windows.  I can not make this
    # work at the moment, so I'm relying on a call to pubsub to do the work.
    def OnChildFocus(self, evt):
        popup = evt.GetEventObject()
        log.debug("OnChildFocus: tab: %s, win=%s new=%s, top=%s" % (self.GetLabel(), popup, evt.GetWindow(), wx.GetApp().GetTopWindow()))
#        if evt.GetWindow() is not None:
#            wx.CallAfter(self.setPopupLoseFocusCallback)

    @property
    def is_shown(self):
        if self.popup is None:
            return False
        popup, child = self.getPopup()
        return popup.IsShown()

    @property
    def managed_window(self):
        if self.popup is None:
            return None
        popup, child = self.getPopup()
        return child

    def setPopupLoseFocusCallback(self):
        """Callback for use within wx.CallAfter to prevent focus being set
        after the control has been removed.
        """
        popup, child = self.getPopup()
        if popup.IsShown():
            log.debug("removing focus from %s" % self.GetLabel())
            self.GetParent().clearRadio()

    def OnLoseChildFocus(self, evt):
        popup = evt.GetEventObject()
        focus = evt.GetWindow()
        log.debug("OnLoseChildFocus: tab: %s, win=%s new=%s, top=%s" % (self.GetLabel(), popup, focus, wx.GetApp().GetTopWindow()))
        if popup is not None:
            if sys.platform == "linux2" or popup != focus:
                wx.CallAfter(self.setPopupLoseFocusCallback)

    def recalc_notification(self):
        popup, child = self.getPopup()
        try:
            count = child.get_notification_count()
            if count != self.notification_count:
                self.notification_count = count
                self.Refresh()
        except AttributeError:
            pass


class SpringTabs(wx.Panel):
    max_popup_width = 800

    def __init__(self, parent, task, *args, **kwargs):
        # Need to remove foreign keyword arguments
        vertical = True
        if 'orientation' in kwargs:
            if kwargs['orientation'] == 'horizontal':
                vertical = False
            del kwargs['orientation']
        popup_direction = 'default'
        if 'popup_direction' in kwargs:
            popup_direction = kwargs['popup_direction']
            del kwargs['popup_direction']
        wx.Panel.__init__(self, parent, *args, **kwargs)
        self.task = task

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self._tabs = []
        if vertical:
            self._tab_renderer = SpringTabItemVerticalRenderer(popup_direction)
        else:
            self._tab_renderer = SpringTabItemHorizontalRenderer(popup_direction)
        self._radio = None

        self._debug_paint = False

        # Using a real wx.PopupWindow seems prevent the focus from being set to
        # the window in the popup.
        if sys.platform == "darwin":
            self._popup_cls = RealPopupWindow
        else:
            self._popup_cls = FakePopupWindow
            self._popup_cls = RealPopupWindow

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)

        self.init_colors()

    def init_colors(self):
        self.notification_background = wx.Colour(240, 120, 120)
        self.notification_brush = wx.Brush(self.notification_background, wx.SOLID)
        self.notification_pen = wx.Pen(self.notification_background, 1, wx.SOLID)
        self.notification_text = wx.Colour(255, 255, 255)
        self.notification_font = wx.Font(8, wx.FONTFAMILY_SWISS, wx.NORMAL, wx.NORMAL)

    def getRenderer(self):
        return self._tab_renderer

    def setRadio(self, item):
        self._processing_radio = True
        if self._radio != item:
            self.clearRadio()
            self.popupItem(item)
        elif not item.GetToggle():
            self.clearRadio()

    def getNewPopup(self):
        popup = self._popup_cls(self)
        return popup

    def popupItem(self, item):
        self._radio = item
        #log.debug("Popping up %s" % item.GetLabel())
        self._tab_renderer.showPopup(self, item)

    def clearRadio(self, data=None):
        # uncomment this to show where the clearRadio is being called
        #import traceback
        #log.debug("".join(traceback.format_stack()))
        print "CLEARRADIO", self._radio
        if self._radio is not None:
            #log.debug("Removing popup %s" % self._radio.GetLabel())
            self._tab_renderer.showPopup(self, self._radio, False)
            self._radio.SetToggle(False, check_popup=False)
        self._radio = None
        self.Refresh()

    def has_popup(self):
        return self._radio is not None

    clear_popup = clearRadio

    def OnPaint(self, evt):
        if self._debug_paint:
            dc = wx.PaintDC(self)

            size = self.GetClientSize()
            dc.SetFont(wx.NORMAL_FONT)
            dc.SetBrush(wx.WHITE_BRUSH)
            dc.SetPen(wx.WHITE_PEN)
            dc.DrawRectangle(0, 0, size.x, size.y)
            dc.SetPen(wx.LIGHT_GREY_PEN)
            dc.DrawLine(0, 0, size.x, size.y)
            dc.DrawLine(0, size.y, size.x, 0)

        #self._tab_renderer.drawTabs(dc, size.x, self._tabs)
        evt.Skip()

    def OnSize(self, evt):
        self.Refresh()
        evt.Skip()

    def addTab(self, title, window_create_callback, window_available_callback=None, **kwargs):
        tab = SpringTabItem(self, label=title, window_cb=window_create_callback,available_cb=window_available_callback, **kwargs)
        self.GetSizer().Add(tab, 0, wx.EXPAND)
        self._tabs.append(tab)
        self.Layout()
        self.Refresh()

    def hasTabs(self):
        return bool(self._tabs)

    def deleteTabs(self):
        self.clearRadio()
        for tab in self._tabs:
            tab.Destroy()
        self._tabs = []

    def update_notifications(self):
        for tab in self._tabs:
            tab.recalc_notification()

    def cancel_popup(self):
        self.task.on_hide_minibuffer_or_cancel(None)


if __name__ == "__main__":
    import wx.calendar
    import wx.stc

    class FontList(wx.Panel):
        def __init__(self, parent, *args, **kwargs):
            wx.Panel.__init__(self, parent, -1)

            e = wx.FontEnumerator()
            e.EnumerateFacenames()
            list = e.GetFacenames()

            list.sort()

            self.lb1 = wx.ListBox(self, -1, wx.DefaultPosition, (200, 250),
                                 list, wx.LB_SINGLE)

            self.Bind(wx.EVT_LISTBOX, self.OnSelect, id=self.lb1.GetId())

            self.txt = wx.StaticText(self, -1, "Sample text...", (285, 50))

            sizer = wx.BoxSizer(wx.VERTICAL)
            sizer.Add(self.txt, 0, wx.EXPAND)
            sizer.Add(self.lb1, 0, wx.EXPAND|wx.TOP, 20)

            self.SetSizer(sizer)
            self.Fit()
            self.Layout()
            log.debug(self.GetSize())

            self.lb1.SetSelection(0)
            self.OnSelect(None)
            wx.FutureCall(300, self.SetTextSize)

        def SetTextSize(self):
            self.txt.SetSize(self.txt.GetBestSize())

        def OnSelect(self, evt):
            face = self.lb1.GetStringSelection()
            font = wx.Font(28, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, face)
            self.txt.SetLabel(face)
            self.txt.SetFont(font)
            if wx.Platform == "__WXMAC__": self.Refresh()

    def ButtonCB(parent, task, **kwargs):
        button = GenToggleButton(parent, -1, "Whatevar!!!")

    def CalendarCB(parent, task, **kwargs):
        wx.calendar.CalendarCtrl(parent, -1, wx.DateTime_Now())

    app = wx.PySimpleApp()
    frm = wx.Frame(None,-1,"Test",style=wx.TAB_TRAVERSAL|wx.DEFAULT_FRAME_STYLE,
                   size=(600,400))
    panel = wx.Panel(frm)
    sizer = wx.BoxSizer(wx.HORIZONTAL)

    # spring tabs for the left side
    tabs1 = SpringTabs(panel, None)
    tabs1.addTab("Calendar", CalendarCB)
    tabs1.addTab("Fonts", FontList)
    tabs1.addTab("Three", ButtonCB)
    sizer.Add(tabs1, 0, wx.EXPAND)

    text = wx.stc.StyledTextCtrl(panel, -1)
    text.SetText("Just a placeholder here.\nThe real action is on the borders.")
    sizer.Add(text, 1, wx.EXPAND)

    # spring tabs for the rigth side
    tabs2 = SpringTabs(panel, None, popup_direction="left")
    tabs2.addTab("Calendar", CalendarCB)
    tabs2.addTab("Five", ButtonCB)
    tabs2.addTab("Six", ButtonCB)
    sizer.Add(tabs2, 0, wx.EXPAND)

    def fixFocus(evt):
        evt.Skip()
    text.Bind(wx.EVT_SET_FOCUS, fixFocus)

    panel.SetAutoLayout(True)
    panel.SetSizer(sizer)
    #sizer.Fit(panel)
    #sizer.SetSizeHints(panel)
    panel.Layout()
    app.SetTopWindow(frm)
    frm.Show()
    app.MainLoop()



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/11/2007
#
#-------------------------------------------------------------------------

""" Traits UI MS Flash editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

if wx.Platform == '__WXMSW__':
    from wx.lib.flashwin import FlashWindow

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

#-------------------------------------------------------------------------
#  '_FlashEditor' class:
#-------------------------------------------------------------------------


class _FlashEditor(Editor):
    """ Traits UI Flash editor.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the table editor is scrollable? This value overrides the default.
    scrollable = True

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = FlashWindow(parent)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        value = self.str_value.strip()
        if value.find('://') < 0:
            value = 'file://' + value

        wx.BeginBusyCursor()
        self.control.LoadMovie(0, value)
        wx.EndBusyCursor()

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for Flash editors:


class FlashEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _FlashEditor
